unit cappo;

interface

uses Classes, Graphics, Controls, Windows, Inifiles, SysUtils,
     Registry;

Type
    TCharSet = set of Char;
    TDirCallback = function(DirName:String):Boolean;

function CappoTokenize(WholeString:String;var StrArray:array of String;Sep:TCharSet):Integer;
function CappoSLTokenize(WholeString:String;SL:TStrings;Sep:TCharSet):Integer;
function CappoStrTokenize(WholeString:String;var StrArray:array of String;Sep:String):Integer;
function CappoQuotesTokenize(WholeString:String;var StrArray:array of String;Sep,Quote:Char):Integer;
procedure ShuffleStrList(var Lst: TStringList);
procedure Squarize(Src,Dest:TCanvas;Squaresize,Height,Width,Speed:Integer;UpdateControl:TControl);
function Power(Base,Exp:Integer):Integer;
function DWORDToHex(DW:DWORD):String;
function DWORDToColourHex(DW:DWORD):String;
function HexToDWORD(Hex:String):DWORD;
function SearchBlock(Data:PChar;DataSize:DWORD;SearchFor:String):Longint; //returns position found at
function TempPath:String;
function WindowsPath:String;
function FileSizeStr(Size:Integer):String;
//returns a string of filesize, using bytes/KB/MB as appropriate.
//Takes 1024 byte standard for filesize
function SwapChar(Src:String;Old,New:Char):String;
function RemoveChar(Src:String;Unwant:Char):String;
function MatchWildcard(Mask:String;Text:String):Boolean;
function MakeColor(R,G,B:Byte):TColor;
procedure GlobalGetMem(var P:PChar;Sz:Integer);
procedure GlobalFreeMem(var P:PChar);
//Allocate PChar memory accessible to any DLL/Application. Use if passing
//dynamic PChar's between libs.
function IntToFullStr(I:Integer):String;
function CountChar(S:String;C:Char):Integer;
function Max(V1,V2:Integer):Integer;
function Min(V1,V2:Integer):Integer;
function MaxF(V1,V2:Single):Single;
function MinF(V1,V2:Single):Single;
function BeginsWith(Big,St:String):Boolean;
function EndsWith(Big,St:String):Boolean;
function AppVersion: Comp;
function ReadVersionInfo: String;
function ReadFileVersionInfo(FName:String): String;

function WalkPath(Start:String;Callback:TDirCallback):Boolean;

function GetFileRec(FName:String;var SR:TSearchRec):Boolean;

Const
     HEXDIGITS = '0123456789ABCDEF';

Type
    TCappoIniFile = class(TIniFile)
                  public
                  procedure ReadSectionNames(Names:TStrings);
    end;

implementation

procedure GlobalGetMem(var P:PChar;Sz:Integer);
var
H:        HGLOBAL;
begin
H := GlobalAlloc(GPTR,Sz);
P := PChar(H);
end;

procedure GlobalFreeMem(var P:PChar);
begin
GlobalFree(HGLOBAL(P));
P := nil;
end;

function CappoTokenize(WholeString:String;var StrArray:array of String;Sep:TCharSet):Integer;
var
I,K:      Integer;
begin
For I:=Low(StrArray) to High(StrArray) do
    StrArray[I]:='';
K:=Low(StrArray)-1;
I:=0;
     While I < Length(WholeString) do begin
            if K=High(StrArray) then Dec(K);
            Inc(I);
            If WholeString[I] in Sep then begin
               Inc(K); StrArray[K]:=StrArray[K]+Copy(WholeString,1,I-1);
               WholeString:=Copy(WholeString,I+1,Length(WholeString));
               I := 0;
            end;
     end;
if (Length(WholeString) > 0) then begin
   If (K <> High(StrArray)) then Inc(K);
   StrArray[K]:=StrArray[K]+WholeString;
end;
Result := K+1;
end;

function CappoSLTokenize(WholeString:String;SL:TStrings;Sep:TCharSet):Integer;
var
I:      Integer;
begin
     SL.Clear;
     I:=0;
     While I < Length(WholeString) do begin
            Inc(I);
            If WholeString[I] in Sep then begin
               SL.Add(Copy(WholeString,1,I-1));
               WholeString:=Copy(WholeString,I+1,Length(WholeString));
               I := 0;
            end;
     end;
if (Length(WholeString) > 0) then begin
   SL.Add(WholeString);
end;
Result := SL.Count;
end;

function CappoQuotesTokenize(WholeString:String;var StrArray:array of String;Sep,Quote:Char):Integer;
var
I,J,K,L,M:      Integer;
begin
For I:=Low(StrArray) to High(StrArray) do
    StrArray[I]:='';
K:=Low(StrArray)-1;
     Repeat begin
            if K=High(StrArray) then Dec(K);
            J := Pos(Quote,WholeString);
            L := Pos(Quote,Copy(WholeString,J+1,Length(WholeString)));
            If (J>0) and (L>0) then M := L+J else M := 1;
            I:=Pos(Sep,Copy(WholeString,M,Length(WholeString)));
            if I > 0 then begin
               Inc(K); StrArray[K]:=StrArray[K]+RemoveChar(Copy(WholeString,1,I+M-2),Quote);
               WholeString:=Copy(WholeString,I+M,Length(WholeString));
            end;
            end;
     until I=0;
if (Length(WholeString) > 0) then begin
   If (K <> High(StrArray)) then Inc(K) else StrArray[K]:=StrArray[K]+Sep;
   StrArray[K]:=StrArray[K]+WholeString;
end;
Result := K+1;
end;


function CappoStrTokenize(WholeString:String;var StrArray:array of String;Sep:String):Integer;
var
I,K:      Integer;
begin
For I:=Low(StrArray) to High(StrArray) do
    StrArray[I]:='';
K:=Low(StrArray)-1;
     Repeat begin
            if K=High(StrArray) then Dec(K);
            I:=Pos(Sep,WholeString);
            if I > 0 then begin
               Inc(K); StrArray[K]:=StrArray[K]+Copy(WholeString,1,I-1);
               WholeString:=Copy(WholeString,I+Length(Sep),Length(WholeString));
            end;
            end;
     until I=0;
if (Length(WholeString) > 0) and (K <> High(StrArray)) then begin
   Inc(K); StrArray[K]:=WholeString;
end;
Result := K+1;
end;

function SwapChar(Src:String;Old,New:Char):String;
var R:   String;
begin
R := Src;
while Pos(Old,R) <> 0 do
      R[Pos(Old,R)] := New;
Result := R;
end;

procedure TCappoIniFile.ReadSectionNames;
var
F:        Textfile;
S,T:      String;
Ar:       array[0..255] of char;
begin
     S := Filename;
     If Not FileExists(S) then begin
        GetWindowsDirectory(Ar,255);
        T := StrPas(Ar); If T[Length(T)] <> '\' then T:=T+'\';
        S := T+S;
        If FileExists(S) then
           AssignFile(F,S)
        else
            Exit;
     end else
     AssignFile(F,S);
     Reset(F);
     while not (eof(F)) do begin
           Readln(F,S);
           If (Length(S) > 0) and (S[1] = '[') and (S[Length(S)] = ']') then
              Names.Add(Copy(S,2,Length(S)-2));
     end;
     CloseFile(F);
end;

procedure ShuffleStrList(var Lst: TStringList);
var
I,J:        Integer;
begin
Randomize;
For I:= Lst.Count downto 1 do begin
    J := Random(I);
    Lst.Exchange(I-1,J);
end;
end;

procedure Squarize(Src,Dest:TCanvas;Squaresize,Height,Width,Speed:Integer;UpdateControl:TControl);
var
Total,I,R,Tmp,Row,Col,X,Y:  Integer;
Order:                  array[1..6000] of Word;
Rct:                    TRect;
begin
Total := ((Width div SQUARESIZE)+1) * ((Height div SQUARESIZE)+1);
//total number of squares
For I:=1 to Total do
    Order[I] := I;
//now shuffle the array
Randomize;
For I:=Total downto 1 do begin
    R := Random(I)+1;
    Tmp := Order[I];
    Order[I] := Order[R]; Order[R] := Tmp;
end;
For I:=1 to Total do begin //now actually start doing the move
    Col := Order[I] mod ((Width div SQUARESIZE)+1); //I mod numpics per row
    Row := Order[I] div ((Width div SQUARESIZE)+1);
    X := Col * SQUARESIZE; Y := Row * SQUARESIZE;
    Rct := Rect(X,Y,X+SQUARESIZE,Y+SQUARESIZE);
    Dest.CopyRect(Rct,Src,Rct);
    If UpdateControl <> nil then UpdateControl.Refresh;
    Sleep (Speed) ; //to give it a bit of time...
end;
end;

function HexToDWORD(Hex:String):DWORD;
var
I:       Integer;
DW:      DWORD;
C:       Char;
begin
DW := 0; Hex := Uppercase(Hex);
For I:=1 to Length(Hex) do begin
    C := Hex[I];
    Case C of
         '0'..'9':   DW := DW + StrToInt(C)* Power(16,Length(Hex)-I);
         'A'..'F':   DW := DW + (Ord(C)-55)* Power(16,Length(Hex)-I);
    end;
end;
HexToDWORD := DW;
end;

function DWORDToHex(DW:DWORD):String;
var
S:       String;
begin
If DW = 0 then S := '0';
while DW <> 0 do begin
      S := HEXDIGITS[(DW and $F)+1] + S;
      DW := DW shr 4;
end;
DWORDToHex := S;
end;

function DWORDToColourHex(DW:DWORD):String;
var
S,S2:    String;
I:       Byte;
begin
S := DWORDToHex(DW and $FF) + DWORDToHex((DW and $FF00) shr 8) + DWORDToHex((DW and $FF0000) shr 16);
S2 := '000000';
For I:=1 to Length(S) do
    S2[Length(S2)-Length(S)+I] := S[I];
DWORDToColourHex := S2;
end;

function Power(Base,Exp:Integer):Integer;
var
I,R:Integer;
begin
     R := 1;
     For I:=1 to Exp do
         R := R * Base;
     Power := R;
end;

function SearchBlock(Data:PChar;DataSize:DWORD;SearchFor:String):Longint; //returns position found at
var
I,J:     Integer;
Found:   Boolean;
begin
Found:=False;
For I:=1 to DataSize do begin
    If (Data[I-1] = SearchFor[1]) and (I+Length(SearchFor)-1 <= DataSize) then begin
       Found := True;
       For J := 2 to Length(SearchFor) do
           If Data[I-2+J] <> SearchFor[J] then Found := False;
       If Found then Break;
    end;
end;
If Found then SearchBlock := I-1 else SearchBlock := -1;
end;

function TempPath:String;
var
NT:      array[0..255] of char;
S:       String;
begin
GetTempPath(255,NT);
S := StrPas(NT);
If S[Length(S)] <> '\' then S:=S+'\';
TempPath := S;
end;

function WindowsPath:String;
var
NT:      array[0..255] of char;
S:       String;
begin
GetWindowsDirectory(NT,255);
S := StrPas(NT);
If S[Length(S)] <> '\' then S:=S+'\';
WindowsPath := S;
end;

function FileSizeStr(Size:Integer):String;
var      R: String;
begin
If Size < 1024 then R := Format('%d bytes',[Size])
else If Size < 102400 then R := Format('%.2n KB',[Size / 1024])
else If Size < (1024*1024) then R := Format('%d KB',[Size div 1024])
else If Size < (1024*1024*1024) then R := Format('%.2n MB',[Size / (1024*1024)])
else R := Format('%.2n GB',[Size / (1024*1024*1024)]);
Result := R;
end;

function MatchWildcard(Mask:String;Text:String):Boolean;
var
I,MPos,TPos:     Integer;
Match:           Boolean;
LookFor:         String;
begin
I := Pos(';',Mask);
If I <> 0 then begin
   Result := MatchWildcard(Copy(Mask,1,I-1),Text) or
             MatchWildcard(Copy(Mask,I+1,Length(Mask)),Text);
   Exit;
end;

MPos := 1; TPos := 1;
Match := True;
While Match do begin
      If (MPos > Length(Mask)) or (TPos > Length(Text)) then Break; 
      Case Mask[MPos] of
      '?':       begin Inc(MPos); Inc(TPos); end;
      '*':       begin
                      If Mpos = Length(Mask) then
                         Break;
                      Lookfor := Copy(Mask,MPos+1,Length(Mask));
                      I := Pos('*',Lookfor);
                      If I<>0 then Lookfor := Copy(Lookfor,1,I-1);
                      I := Pos('?',Lookfor);
                      If I<>0 then Lookfor := Copy(Lookfor,1,I-1);
                      While (Copy(Text,TPos,Length(Lookfor)) <> Lookfor) and (Tpos <= Length(Text)) do
                            Inc(TPos);
                      If (TPos+Length(Lookfor)-1) > Length(Text) then begin
                         Match := False;
                         Break;
                      end;
                      Inc(MPos,1); //Inc(TPos);
                 end;
      Else       If Mask[MPos] = Text[TPos] then begin
                    Inc(MPos); Inc(TPos);
                 end else begin
                     Match := False;
                     Break;
                 end;
      end;
      If (MPos > Length(Mask)) and (TPos <= Length(Text)) then Match := False;
      If (TPos > Length(Text)) and (Mpos <= Length(Mask)) then 
         If (Mpos=Length(Mask)) and (Mask[MPos] = '*') then {we're OK}
         else Match := False;
end;
Result := Match;
end;

function MakeColor(R,G,B:Byte):TColor;
begin
Result := TColor(R or (G shl 8) or (B shl 16) );
end;

function RemoveChar(Src:String;Unwant:Char):String;
begin
While Pos(Unwant,Src) > 0 do
      Delete(Src,Pos(Unwant,Src),1);
Result := Src;
end;

function IntToFullStr(I:Integer):String;
begin
If I<0 then Result := IntToStr(I) else Result := '+' + IntToStr(I);
end;

function CountChar(S:String;C:Char):Integer;
var
I,Cnt:   Integer;
begin
Cnt := 0;
For I := 1 to Length(S) do
    If S[I]=C then Inc(Cnt);
Result := Cnt;
end;

function Max(V1,V2:Integer):Integer;
begin
If V1>V2 then Result := V1 else Result := V2;
end;

function Min(V1,V2:Integer):Integer;
begin
If V1<V2 then Result := V1 else Result := V2;
end;

function MaxF(V1,V2:Single):Single;
begin
If V1>V2 then Result := V1 else Result := V2;
end;

function MinF(V1,V2:Single):Single;
begin
If V1<V2 then Result := V1 else Result := V2;
end;

function BeginsWith(Big,St:String):Boolean;
begin
Result := ( Copy(Big,1,Length(St)) = St);
end;

function EndsWith(Big,St:String):Boolean;
begin
Result := ( Copy(Big,Length(Big)-Length(St)+1,Length(Big)) = St);
end;

function WalkPath(Start:String;Callback:TDirCallback):Boolean;
var
SR:     TSearchRec;
I:      Integer;
begin
Result := False;
If Not Callback(Start) then Exit;
Result := True;
I := FindFirst(Start+'*.*',faDirectory,SR);
While I=0 do begin
  If (SR.Attr and faDirectory)=faDirectory then
    If Not WalkPath(Start+SR.Name+'\',Callback) then begin
      Result := False;
      Exit;
    end;
  I := FindNext(SR);
end;
FindClose(SR);
end;

function GetFileRec(FName:String;var SR:TSearchRec):Boolean;
begin
Result := FindFirst(FName,0,SR) = 0;
FindClose(SR);
end;

function AppVersion: Comp;
type
  T2Words = record
    case Boolean of
      TRUE: (C: Cardinal);
      FALSE: (Lo, Hi: Word);
  end;
var
  size:   Cardinal;
  p, vs:  Pointer;
  PCmp:   ^Comp;
  NT:     Array[0..255] of Char;
begin
  GetModuleFilename(hInstance,@NT,255);
  size := GetFileVersionInfoSize(NT, DWORD(size));
  if size > 0 then
  begin
    GetMem(p, size);
    GetFileVersionInfo(NT, 0, size, p);
    VerQueryValue(p, '\', vs, DWORD(size));
    with TVSFixedFileInfo(vs^) do
    begin
      PCmp := @dwFileVersionMS;
      Result := PCmp^;
    end;
    FreeMem(p);
  end else Result := 0;
end;

function ReadFileVersionInfo(FName:String): String;
type
  T2Words = record
    case Boolean of
      TRUE: (C: Cardinal);
      FALSE: (Lo, Hi: Word);
  end;
var
  size: Cardinal;
  p, vs: Pointer;
  MS, LS: T2Words;
  NT:     Array[0..255] of Char;
begin
  StrPCopy(NT,FName);
  size := GetFileVersionInfoSize(NT, DWORD(size));
  if size > 0 then
  begin
    GetMem(p, size);
    GetFileVersionInfo(NT, 0, size, p);
    VerQueryValue(p, '\', vs, DWORD(size));
    with TVSFixedFileInfo(vs^) do
    begin
      MS.C := dwFileVersionMS;
      LS.C := dwFileVersionLS;
      Result := Format('%d.%d.%d build %d', [MS.Hi, MS.Lo, LS.Hi, LS.Lo]);
    end;
    FreeMem(p);
  end
  else Result := 'unknown';

end;

function ReadVersionInfo: String;
type
  T2Words = record
    case Boolean of
      TRUE: (C: Cardinal);
      FALSE: (Lo, Hi: Word);
  end;
var
  size: Cardinal;
  p, vs: Pointer;
  MS, LS: T2Words;
  NT:     Array[0..255] of Char;
begin
  GetModuleFilename(hInstance,@NT,255);
  size := GetFileVersionInfoSize(NT, DWORD(size));
  if size > 0 then
  begin
    GetMem(p, size);
    GetFileVersionInfo(NT, 0, size, p);
    VerQueryValue(p, '\', vs, DWORD(size));
    with TVSFixedFileInfo(vs^) do
    begin
      MS.C := dwFileVersionMS;
      LS.C := dwFileVersionLS;
      Result := Format('%d.%d.%d build %d', [MS.Hi, MS.Lo, LS.Hi, LS.Lo]);
    end;
    FreeMem(p);
  end
  else Result := 'unknown';

end;

end.
