unit DebugLog;

// DEFINE AUTOINIT if you want it to initialise automatically
// DEFINE DEBUG or nothing will happen!

interface

{$IFDEF DEBUG}
uses Classes, SysUtils, Windows;

var
  LogFilename:    String;
{$ELSE}
uses Classes;
{$ENDIF}

procedure Log(Msg:String);
procedure LogFmt(const Fmt:String; const Args: Array of const);
procedure InitLog;
procedure DumpLog(Filename:String);
procedure AssignDebugMessages(Tgt: TStrings);

implementation

{$IFDEF DEBUG}

var
  Msgs:     TStringList;

procedure Log(Msg:String);
begin
If Msgs=nil then
{$IFDEF AUTOINIT}
InitLog;
{$ELSE}
Exit;
{$ENDIF}
Msgs.Add(Msg);
end;

procedure LogFmt(const Fmt:String; const Args: Array of const);
begin
If Msgs=nil then
{$IFDEF AUTOINIT}
InitLog;
{$ELSE}
Exit;
{$ENDIF}
Msgs.Add(Format(Fmt,Args));
end;

procedure InitLog;
var
NT:     Array[0..MAX_PATH] of Char;
Fil:    String;
begin
If Msgs<>nil then Exit;
GetModuleFilename(hInstance,@NT,255);
Fil := ExtractFilePath(NT);
If Fil[Length(Fil)] <> '\' then Fil := Fil + '\';
If Not FileExists(Fil+'nul') then Fil := ExtractFilePath(ParamStr(0));
Fil := Fil + 'LOG.TXT';
LogFilename := Fil;
Msgs := TStringList.Create;
Log('Debug log created at '+DateTimeToStr(Now)+' for program '+ParamStr(0));
end;

procedure DumpLog(Filename:String);
begin
Msgs.SaveToFile(Filename);
end;

procedure AssignDebugMessages(Tgt: TStrings);
begin
Tgt.Assign(Msgs);
end;

initialization
{$IFDEF AUTOINIT}
InitLog;
{$ENDIF}
finalization
DumpLog(LogFilename);
Msgs.Free;

{$ELSE}

procedure Log(Msg:ShortString);
begin end;

procedure LogFmt(const Fmt:String; const Args: Array of const);
begin end;
procedure InitLog;
begin end;

procedure DumpLog(Filename:String);
begin end;

procedure AssignDebugMessages(Tgt: TStrings);
begin end;

{$ENDIF}
end.
