unit ff7ed;

{
FF7 Editor functions.


}

interface

Uses Windows,Classes,SysUtils,FFLZS,DebugLog,Graphics,PluginTypes,FF7Types,
      BaseUtil;

Type
  TBackgroundDecodeThread = class(TThread)
    private
      Source:       TMemoryStream;
    public
      Output:       TBitmap;
      Constructor Create(Src:TMemoryStream);
      Destructor Destroy; Override;
      procedure Execute; Override;
      property ReturnValue;
  end;
  TExtractProc = function(Src:TMemoryStream;var Table:TList): TMemoryStream;

Const
  ValidChars : set of Char = ['a'..'z','A'..'Z',' '..'@',#0,'_',#129];

function ExtractTextRT(Src:TMemoryStream;var Table:TList): TMemoryStream;
function ExtractTextRTN(Src:TMemoryStream;var Table:TList;var Info:TTextInfoRec): TMemoryStream;

function FF7DecodeText(Src:PByte;SrcLen:Integer): String;
function FF7DecodeTextS(Src:TStream): String;
//function FF7DecodeTextMonitorS(Src:TStream;ByteRef:TStream): String;
function FF7EncodeTextS(Src:String): TMemoryStream;
function FF7EncodeTextQS(Src:String): TMemoryStream;

function GetFF7Background(Src:TMemoryStream):TBitmap;

function TextSizeNeeded(Txt:String):TPoint;
function TextBlockSizeNeeded(Name,Txt:String;Speech:Boolean):TPoint;

var
ChPC: Array[0..255] of Char;
StPC: Array[0..255] of String[23];
InSt: Set of Char;
TEXT_FACTOR: Integer = 13;

TEXT_OFFSET: Integer = -1;
TEXT_SIZE: Integer = -1;
TABLE_START: Integer = -1;
TEXT_FILE_OFFSET: Integer = -1;

implementation

var
  Text_Sizes: Array [0..115] of Integer =
   (10,06,10,20,14,20,18,06,10,10,14,
    14,08,10,06,14,16,10,16,16,16,
    16,16,16,16,16,06,08,14,16,14,12,
    20,18,14,16,16,14,14,16,16,06,
    12,14,14,22,16,18,14,18,14,14,14,
    16,18,22,16,18,14,08,12,08,14,
    16,08,14,14,12,14,14,12,14,14,06,
    08,12,06,22,14,14,14,14,10,12,
    12,14,14,22,14,16,12,10,06,10,16,
    12,00,00,00,00,00,00,00,00,00,
    00,00,00,00,00,00,00,00,00,00,12
    );//subtract one for optimal view...

function FF7DecodeText(Src:PByte;SrcLen:Integer): String;
var
I:      Integer;
begin
Result := '';
//SetLength(Result,SrcLen);
For I := 1 to SrcLen do begin
  Result := Result + StPc[Src^];
  Inc(Src);
end;
end;

function FF7DecodeTextS(Src:TStream): String;
var
I:      Integer;
B:      Byte;
begin
Result := '';
Src.Position := 0;
//SetLength(Result,SrcLen);
For I := 1 to Src.Size do begin
  Src.ReadBuffer(B,1);
  Result := Result + StPc[B];
end;
end;

function FF7DecodeTextMonitorS(Src:TStream;ByteRef:TStream): String;
var
I,J:    Integer;
B:      Byte;
BA:     Array[0..31] of Integer;
begin
Result := '';
//SetLength(Result,SrcLen);
For I := 0 to Src.Size-1 do begin
  Src.ReadBuffer(B,1);
  Result := Result + StPc[B];
  For J := 0 to 31 do BA[J] := I;
  ByteRef.WriteBuffer(BA,4*Length(StPc[B]));
end;
I := Src.Size;
ByteRef.WriteBuffer(I,4);
end;

function FF7EncodeTextQS(Src:String): TMemoryStream;
var
DstLen,Lng,Cur:  Integer;
I,B:            Byte;
begin
Result := TMemoryStream.Create;
Result.Size := Length(Src) div 2;
DstLen := 0;
While Src <> '' do begin
  Lng := 1;
  If Src[1]='[' then begin
    Lng := 0; Cur := 0;
    For I := 0 to 255 do
      If BeginsWith(Src,StPc[I]) and (Length(StPc[I]) > Lng) then begin
        Cur := I; Lng := Length(StPc[I]);
      end;
      B := Cur;
  end else
    B := Byte(FF7Ch[Byte(Src[1])]);
  Result.WriteBuffer(B,1); Inc(DstLen);
  Delete(Src,1,Lng);
end;
B := 255;
Result.WriteBuffer(B,1); Inc(DstLen);
Result.Size := DstLen;
end;

function FF7EncodeTextS(Src:String): TMemoryStream;
var
DstLen,Lng,Cur:  Integer;
I,B:            Byte;
begin
Result := TMemoryStream.Create;
Result.Size := Length(Src) div 2;
DstLen := 0;
While Src <> '' do begin
  Lng := 1;
  If Src[1] in InSt then begin
    Lng := 0; Cur := 0;
    For I := 0 to 255 do
      If BeginsWith(Src,StPc[I]) and (Length(StPc[I]) > Lng) then begin
        Cur := I; Lng := Length(StPc[I]);
      end;
      B := Cur;
  end else
    B := Byte(FF7Ch[Byte(Src[1])]);
  Result.WriteBuffer(B,1); Inc(DstLen);
  Delete(Src,1,Lng);
end;
Result.Size := DstLen;
end;

function ExtractTextRT(Src:TMemoryStream;var Table:TList): TMemoryStream;
var
Len,Num,I:    Integer;
Null,Valid:   Boolean;
PI:           Integer;
Offsets:      TList;
Nme:          TFF7Name;
PB:           Byte;
Cur:          Byte;
Tmp:          TMemoryStream;
Last,Cnt,Fac: Array[0..1] of Byte;
W:            Word;
begin
Src.Position := 2;
Src.ReadBuffer(Num,4);
Log('Reading TOC...');
Offsets := TList.Create;
Tmp := TMemoryStream.Create;
try
Result := TMemoryStream.Create;
If Num<1 then Exit;
For I := 1 to Num do begin
  Src.ReadBuffer(PI,4);
  Offsets.Add(Pointer(PI));
end;
Src.Position := Integer(Offsets[0]);
Src.Position := Src.Position + 20;
Log('Reading 8-strings...');
Repeat
  Src.ReadBuffer(Nme,8);
  Valid := True;
  Null := False;
  For I := 0 to 7 do begin
    Null := Null or (Nme[I]=#0);
    Valid := Valid and ( (Nme[I] in ValidChars) and ( (Not Null) or (Nme[I]=#0) ) );
  end;
Until Not Valid;
Src.Seek(-8,soFromCurrent);
TEXT_FILE_OFFSET := Src.Position;
Src.ReadBuffer(PI,4);
Tmp.Size := PI;
Src.Position := PI + Integer(Offsets[0]);
Len := 0;
Last[0] := 0; Last[1] := 0;
Cnt := Last; Fac := Cnt;
Log('Ready to start byte scan...');
Cur := 0;
While (Src.Position >0) and (Src.Position < Src.Size) do begin
  Src.ReadBuffer(PB,1);//!!!!
  Src.Seek(-2,soFromCurrent);
  Cur := Src.Position mod 2;
  If PB = Last[Cur] then begin Inc(Cnt[Cur]); If (Fac[Cur] mod 10)<>9 then Inc(Fac[Cur]);
    end else If PB = Last[Cur]-1 then begin Inc(Cnt[Cur],10); Inc(Fac[Cur],10); end
      else begin Cnt[Cur] := 0; Fac[Cur] := 0; end;
    Last[Cur] := PB;
  If (Fac[Cur]>TEXT_FACTOR)  and ((Fac[Cur] mod 10)>1) then begin
    Dec(Len,((Cnt[Cur] mod 10)+(Cnt[Cur] div 10))*2);
    Break;
  end;
  Tmp.WriteBuffer(PB,1);
  Inc(Len);
end;
TEXT_OFFSET := (PI + Integer(Offsets[0])) - Len + 1;
TEXT_SIZE := Len;
Log('Broke from search at decompressed position '+IntToStr(Src.Position)+' of Count '+IntToStr(Cnt[Cur]));
Src.Position := PI + Integer(Offsets[0]);
Src.Seek(-(Len-1),soFromCurrent);
Result := TMemoryStream.Create;
Result.Position := 0;
Result.CopyFrom(Src,Len);
Src.Position := PI + Integer(Offsets[0]);
Src.Seek(-(Len+1),soFromCurrent);
Table := TList.Create;
If (Src.Position>Src.Size) or (Src.Position < 0) then Src.Position := 2;
Repeat
  Src.ReadBuffer(W,2);
  Src.Seek(-4,soFromCurrent); //!!!!
  Table.Insert(0,Pointer(W));
  If Table.Count>2 then
    If (Integer(Table[2]) - 2*(Table.Count-1))=0 then Break;
Until (W=0) or (Src.Position < 2);
Table.Delete(0);
TABLE_START := Src.Position+4;
For I := 0 to Table.Count-1 do begin
  Num := Word(Table[I]);
//  Dec(Num,4);
  Dec(Num,2*Table.Count);
  Table[I] := Pointer(Num);
end;
finally
Offsets.Free;
Tmp.Free;
Log('Finished ExtractRawText');
end;
end;

function ExtractTextRTN(Src:TMemoryStream;var Table:TList;var Info:TTextInfoRec): TMemoryStream;
var
Len,Num,I:    Integer;
Null,Valid:   Boolean;
PI:           Integer;
Offsets:      TList;
Nme:          TFF7Name;
PB:           Byte;
Cur:          Byte;
Tmp:          TMemoryStream;
Last,Cnt,Fac: Array[0..1] of Byte;
LastW:        Array[0..1] of Word;
AkOff,EOD:    Integer;
W:            Word;
begin
Src.Position := 2;
Src.ReadBuffer(Num,4);
Log('Reading TOC...');
Offsets := TList.Create;
Tmp := TMemoryStream.Create;
try
Result := TMemoryStream.Create;
Table := TList.Create;
If Num<1 then Exit;
For I := 1 to Num do begin
  Src.ReadBuffer(PI,4);
  Offsets.Add(Pointer(PI));
end;
Src.Position := Integer(Offsets[0]);
Src.Position := Src.Position + 20;
Log('Reading 8-strings...');
Repeat
  Src.ReadBuffer(Nme,8);
  Valid := True;
  Null := False;
  For I := 0 to 7 do begin
    Null := Null or (Nme[I]=#0);
    Valid := Valid and ( (Nme[I] in ValidChars) and ( (Not Null) or (Nme[I]=#0) ) );
  end;
Until Not Valid;
Src.Seek(-8,soFromCurrent);
Info.FileOffset := Src.Position;
Src.ReadBuffer(PI,4);
If (PI<5000000) then Tmp.Size := PI;
Src.Position := PI + Integer(Offsets[0]);
If (Src.Position>=Src.Size) or (Src.Position<0) then begin
  Offsets[0] := Offsets[1];
  PI := -1;
  Src.Position := PI + Integer(Offsets[0]);
  Info.UsedScriptOffset := False;
end else Info.UsedScriptOffset := True;
If Src.Position>=Src.Size then Exit;
EOD := 0;
While (EOD <> AKAO_CODE) and (Src.Position < (PI + Integer(Offsets[0]) + 128) ) do begin
  Src.ReadBuffer(EOD,4);
  Src.Seek(-3,soFromCurrent);
end;
If EOD=AKAO_CODE then PI := Src.Position-6-Integer(Offsets[0]);

Src.Position := PI + Integer(Offsets[0]);
PB := 1;
For I := 1 to 8 do begin
  Src.ReadBuffer(PB,1);
  If (PB=0) or (PB=$FF) then begin
    Dec(PI); Src.Seek(-2,soFromCurrent);
  end else begin
    Src.Seek(-1,soFromCurrent); Break;
  end;
end;
Len := 0;
Last[0] := 0; Last[1] := 0;
Cnt := Last; Fac := Cnt;
Log('Ready to start byte scan...');
Cur := 0; EOD := 0;
While (Src.Position >0) and (Src.Position < Src.Size) do begin
  Src.ReadBuffer(PB,1);//!!!!
  Src.Seek(-2,soFromCurrent);
{  If StPC[PB]=Char(NA) then begin
    PB := 0;
    Break;
  end;}
  If (EOD>0) and (PB=$FF) and (Src.Position-EOD < -3) then begin
    EOD := Src.Position-EOD;
    AkOff := Src.Position;
  end;
  If (EOD=0) and (PB=$FF) then EOD := Src.Position;
  Cur := Src.Position mod 2;
  W := PB or (Last[1-Cur] shl 8);
  If PB = Last[Cur] then begin Inc(Cnt[Cur]); If (Fac[Cur] mod 10)<>9 then Inc(Fac[Cur]);
    end else If PB = Last[Cur]-1 then begin Inc(Cnt[Cur],10); Inc(Fac[Cur],10); end
      else begin Cnt[Cur] := 0; Fac[Cur] := 0; end;
  Last[Cur] := PB;
  If EOD<0 then
    If (LastW[Cur]-W = -EOD) and (Abs(Src.Position+W-AkOff)<$FF) then begin
      Inc(Fac[Cur],22);
      Inc(Cnt[Cur],1);
      Dec(Len);
    end;
{  If (W=0) and (LastW[Cur]>$FF) then begin
    Inc(Fac[Cur],22);
//    Dec(Len);
//    Inc(Cnt[Cur],2);
  end;}
  If (Fac[Cur]>Info.Factor)  and ((Fac[Cur] mod 10)>1) then begin
    Dec(Len,((Cnt[Cur] mod 10)+(Cnt[Cur] div 10))*2);
    Break;
  end;
  LastW[Cur] := W;
  Tmp.WriteBuffer(PB,1);
  Inc(Len);
end;
Info.Offset := (PI + Integer(Offsets[0])) - Len + 1;
Info.Size := Len;
Log('Broke from search at decompressed position '+IntToStr(Src.Position)+' of Count '+IntToStr(Fac[Cur]));
Src.Position := PI + Integer(Offsets[0]);
Src.Seek(-(Len-1),soFromCurrent);
Result := TMemoryStream.Create;
Result.Position := 0;
Result.CopyFrom(Src,Len);
Src.Position := PI + Integer(Offsets[0]);
Src.Seek(-(Len+1),soFromCurrent);
If (Src.Position>Src.Size) or (Src.Position < 0) then Src.Position := 2;
Repeat
  Src.ReadBuffer(W,2);
  Src.Seek(-4,soFromCurrent); //!!!!
  Table.Insert(0,Pointer(W));
  If Table.Count>2 then
    If (Integer(Table[2]) - 2*(Table.Count-1))=0 then Break;
Until (W=0) or (Src.Position < 2);
Table.Delete(0);
Info.Start := Src.Position+4;
For I := 0 to Table.Count-1 do begin
  Num := Word(Table[I]);
//  Dec(Num,4);
  Dec(Num,2*Table.Count);
  Table[I] := Pointer(Num);
end;
finally
Offsets.Free;
Tmp.Free;
Log('Finished ExtractRawText');
end;
end;

function GetFF7Background(Src:TMemoryStream):TBitmap;
var
K,J,TOffset,TDest,I,B3Off,BOff:               Integer;
DCol,bgnsprites2,bgwidth,bgheight,bgnsprites: Word;
bgpsprites2,bgpsprites,psprite:               PFF7BGSPRITE;
Col,Dest,Palette:                             PWord;
Source,Background,Picture,PB:                 PByte;
Bmp: TBitmap;
begin
//Src := LZSFileDecompressS('D:\games\ff7\data\field\gldst.dec');
Src.Position := 2 + (8+1)*4;
Src.ReadBuffer(BOff,4); Inc(BOff,4);
Src.Position := 2 + (3+1)*4;
Src.ReadBuffer(B3Off,4); Inc(B3Off,4);
Src.Position := BOff + $28;
Src.ReadBuffer(bgwidth,2);
Src.Position := BOff + $2A;
Src.ReadBuffer(bgheight,2);
Src.Position := BOff + $2C;
Src.ReadBuffer(bgnsprites,2);
GetMem(bgpsprites,Sizeof(TFF7BGSPRITE)*bgnsprites);
Src.Position := BOff + $32;
Src.ReadBuffer(bgpsprites^,Sizeof(TFF7BGSPRITE)*bgnsprites);
Src.Position := BOff + bgnsprites*52+$39;
Src.ReadBuffer(bgnsprites2,2);
GetMem(bgpsprites2,Sizeof(TFF7BGSPRITE)*bgnsprites2);
Src.Position := BOff + $32 + bgnsprites*52 + $1B;
Src.ReadBuffer(bgpsprites2^,Sizeof(TFF7BGSPRITE)*bgnsprites2);
PB := Src.Memory;
Inc(PB,B3Off + $C);
Palette := Pointer(PB);
Picture := Src.Memory;
Inc(Picture,BOff + bgnsprites*52 + bgnsprites2*52 + $58);
GetMem(Background,bgwidth*bgheight*2);
ZeroMemory(Background,bgwidth*bgheight*2);

For I := 0 to bgnsprites-1 do begin
  PSprite := BGPSprites;
  Inc(PSprite,I);
  TOffset := (PSprite^.Page shl 16) or ( (PSprite^.SrcY shl 8) or (PSprite^.SrcX) ) + (PSprite^.Page+1)*6;
  TDest := ((PSprite^.Y + (bgheight shr 1))*bgwidth)+(PSprite^.X+(bgwidth shr 1));
  TDest := TDest shl 1;
  If PSprite^.Sfx <> 0 then Continue;
  For J := 0 to 15 do begin
    Source := Picture;
    If TOffset > Src.Size then Continue;
    Inc(Source,TOffset);
    PB := Background;
    If TDest > (BGWidth*BGHeight*2) then Continue;
    Inc(PB,TDest);
    Dest := Pointer(PB);
    For K := 0 to 15 do begin
      If Source^=0 then begin
        Inc(Source); Inc(Dest); Continue;
      end;
      If (Integer(Dest) > ( Integer(Background) + 2*BGHeight*BGWidth - 2 )) or (Integer(Dest) < Integer(Background)) then Continue;
      Col := Palette;
      Inc(Col, (PSprite^.Pal shl 8) + Source^ );
      If Integer(Col) > ( Integer(Src.Memory) + Src.Size ) then Continue;
      Inc(Source);
      DCol := 0;
      DCol := DCol or ( (Col^ and $1F) shl 10 );
      DCol := DCol or (Col^ and $3E0);
      DCol := DCol or ( (Col^ and $7C00) shr 10);
      Dest^ := DCol;
      Inc(Dest);
    end;
    TDest := TDest + (BGWidth shl 1);
    Inc(TOffset,256);
  end;
end;

For I := 0 to bgnsprites2-1 do begin
  PSprite := BGPSprites2;
  Inc(PSprite,I);
  TOffset := (PSprite^.Page shl 16) or ( (PSprite^.SrcY shl 8) or (PSprite^.SrcX) ) + (PSprite^.Page+1)*6;
  TDest := ((PSprite^.Y + (bgheight shr 1))*bgwidth)+(PSprite^.X+(bgwidth shr 1));
  TDest := TDest shl 1;
  If PSprite^.Sfx <> 0 then Continue;
  For J := 0 to 15 do begin
    Source := Picture;
    If TOffset > Src.Size then Continue;
    Inc(Source,TOffset);
    PB := Background;
    If TDest > (BGWidth*BGHeight*2) then Continue;
    Inc(PB,TDest);
    Dest := Pointer(PB);
    For K := 0 to 15 do begin
      If Source^=0 then begin
        Inc(Source); Inc(Dest); Continue;
      end;
      If (Integer(Dest) > ( Integer(Background) + 2*BGHeight*BGWidth - 2 )) or (Integer(Dest) < Integer(Background)) then Continue;
      Col := Palette;
      Inc(Col, (PSprite^.Pal shl 8) + Source^ );
      Inc(Source);
      DCol := 0;
      DCol := DCol or ( (Col^ and $1F) shl 10 );
      DCol := DCol or (Col^ and $3E0);
      DCol := DCol or ( (Col^ and $7C00) shr 10);
      Dest^ := DCol; //!!!
      Inc(Dest);
    end;
    TDest := TDest + (BGWidth shl 1);
    Inc(TOffset,256);
  end;
end;

Bmp := TBitmap.Create;
Bmp.PixelFormat := pf15Bit;
Bmp.Width := BGWidth;
Bmp.Height := BGHeight;
PB := Background;
For I := 0 to BGHeight-1 do begin
  Source := Bmp.ScanLine[I];
  CopyMemory(Source,PB,BGWidth*2);
  Inc(PB,BGWidth*2);
end;
FreeMem(Background);
FreeMem(BGPSprites);
FreeMem(BGPSprites2);
Result := Bmp;
end;

Constructor TBackgroundDecodeThread.Create(Src:TMemoryStream);
begin
Inherited Create(True);
Source := TMemoryStream.Create;
Source.CopyFrom(Src,0);
Output := TBitmap.Create;
end;

Destructor TBackgroundDecodeThread.Destroy;
begin
Source.Free;
Inherited;
end;

procedure TBackgroundDecodeThread.Execute;
var
K,J,TOffset,TDest,I,B3Off,BOff:               Integer;
DCol,bgnsprites2,bgwidth,bgheight,bgnsprites: Word;
bgpsprites2,bgpsprites,psprite:               PFF7BGSPRITE;
Col,Dest,Palette:                             PWord;
PSrc,Background,Picture,PB:                   PByte;
Step:                                         TSteppedLZS;
DCmp:                                         TMemoryStream;
begin

ReturnValue := 0;

Step := TSteppedLZS.Create;
Step.Decompress(Source);

While (Not Terminated) and (Not Step.Done) do
  Step.Step;
If Terminated then Exit;

DCmp := TMemoryStream.Create;
DCmp.CopyFrom(Step.Dst,0);
Step.Free;
Try

DCmp.Position := 2 + (8+1)*4;
DCmp.ReadBuffer(BOff,4); Inc(BOff,4);
DCmp.Position := 2 + (3+1)*4;
DCmp.ReadBuffer(B3Off,4); Inc(B3Off,4);
DCmp.Position := BOff + $28;
DCmp.ReadBuffer(bgwidth,2);
DCmp.Position := BOff + $2A;
DCmp.ReadBuffer(bgheight,2);
DCmp.Position := BOff + $2C;
DCmp.ReadBuffer(bgnsprites,2);
GetMem(bgpsprites,Sizeof(TFF7BGSPRITE)*bgnsprites);
DCmp.Position := BOff + $32;
DCmp.ReadBuffer(bgpsprites^,Sizeof(TFF7BGSPRITE)*bgnsprites);
DCmp.Position := BOff + bgnsprites*52+$39;
DCmp.ReadBuffer(bgnsprites2,2);
GetMem(bgpsprites2,Sizeof(TFF7BGSPRITE)*bgnsprites2);
DCmp.Position := BOff + $32 + bgnsprites*52 + $1B;
DCmp.ReadBuffer(bgpsprites2^,Sizeof(TFF7BGSPRITE)*bgnsprites2);
PB := DCmp.Memory;
Inc(PB,B3Off + $C);
Palette := Pointer(PB);
Picture := DCmp.Memory;
Inc(Picture,BOff + bgnsprites*52 + bgnsprites2*52 + $58);
GetMem(Background,bgwidth*bgheight*2);
ZeroMemory(Background,bgwidth*bgheight*2);
If Terminated then Exit;

For I := 0 to bgnsprites-1 do begin
  PSprite := BGPSprites;
  Inc(PSprite,I);
  TOffset := (PSprite^.Page shl 16) or ( (PSprite^.SrcY shl 8) or (PSprite^.SrcX) ) + (PSprite^.Page+1)*6;
  TDest := ((PSprite^.Y + (bgheight shr 1))*bgwidth)+(PSprite^.X+(bgwidth shr 1));
  TDest := TDest shl 1;
  If PSprite^.Sfx <> 0 then Continue;
  For J := 0 to 15 do begin
    PSrc := Picture;
    If TOffset > DCmp.Size then Continue;
    Inc(PSrc,TOffset);
    PB := Background;
    If TDest > (BGWidth*BGHeight*2) then Continue;
    Inc(PB,TDest);
    Dest := Pointer(PB);
    For K := 0 to 15 do begin
      If PSrc^=0 then begin
        Inc(PSrc); Inc(Dest); Continue;
      end;
      If Integer(Dest) > ( Integer(Background) + 2*BGHeight*BGWidth ) then Continue;
      Col := Palette;
      Inc(Col, (PSprite^.Pal shl 8) + PSrc^ );
      If Integer(Col) > ( Integer(DCmp.Memory) + DCmp.Size ) then Continue;
      Inc(PSrc);
      DCol := 0;
      DCol := DCol or ( (Col^ and $1F) shl 10 );
      DCol := DCol or (Col^ and $3E0);
      DCol := DCol or ( (Col^ and $7C00) shr 10);
      Dest^ := DCol;
      Inc(Dest);
    end;
    If Terminated then Exit;
    TDest := TDest + (BGWidth shl 1);
    Inc(TOffset,256);
  end;
  If Terminated then Exit;
end;

For I := 0 to bgnsprites2-1 do begin
  PSprite := BGPSprites2;
  Inc(PSprite,I);
  TOffset := (PSprite^.Page shl 16) or ( (PSprite^.SrcY shl 8) or (PSprite^.SrcX) ) + (PSprite^.Page+1)*6;
  TDest := ((PSprite^.Y + (bgheight shr 1))*bgwidth)+(PSprite^.X+(bgwidth shr 1));
  TDest := TDest shl 1;
  If PSprite^.Sfx <> 0 then Continue;
  For J := 0 to 15 do begin
    PSrc := Picture;
    If TOffset > DCmp.Size then Continue;
    Inc(PSrc,TOffset);
    PB := Background;
    If TDest > (BGWidth*BGHeight*2) then Continue;
    Inc(PB,TDest);
    Dest := Pointer(PB);
    For K := 0 to 15 do begin
      If PSrc^=0 then begin
        Inc(PSrc); Inc(Dest); Continue;
      end;
      Col := Palette;
      Inc(Col, (PSprite^.Pal shl 8) + PSrc^ );
      Inc(PSrc);
      DCol := 0;
      DCol := DCol or ( (Col^ and $1F) shl 10 );
      DCol := DCol or (Col^ and $3E0);
      DCol := DCol or ( (Col^ and $7C00) shr 10);
      Dest^ := DCol; //!!!
      Inc(Dest);
    end;
    If Terminated then Exit;
    TDest := TDest + (BGWidth shl 1);
    Inc(TOffset,256);
  end;
  If Terminated then Exit;
end;

Output.PixelFormat := pf15Bit;
Output.Width := BGWidth;
Output.Height := BGHeight;
PB := Background;
For I := 0 to BGHeight-1 do begin
  PSrc := Output.ScanLine[I];
  CopyMemory(PSrc,PB,BGWidth*2);
  Inc(PB,BGWidth*2);
  If Terminated then Exit;
end;
finally
FreeMem(Background);
FreeMem(BGPSprites);
FreeMem(BGPSprites2);
DCmp.Free;
end;
ReturnValue := 1;
end;


function MaxPoint(P1,P2:TPoint):TPoint;
begin
Result.X := Max(P1.X,P2.X);
Result.Y := Max(P1.Y,P2.Y);
end;

function TextSizeNeeded(Txt:String):TPoint;
var
X,Y:  Integer;
I:    Byte;
begin
X := 15; Y := 20;
Result := Point(0,0);

While Txt<>'' do begin
  If BeginsWith(Txt,'[Colour:White]') then begin
    Delete(Txt,1,14); Continue;
  end;
  If BeginsWith(Txt,'[Colour:Cyan]') then begin
    Delete(Txt,1,13); Continue;
  end;
  If BeginsWith(Txt,'[Colour:Green]') then begin
    Delete(Txt,1,14); Continue;
  end;
  If BeginsWith(Txt,'[Colour:Purple]') then begin
    Delete(Txt,1,15); Continue;
  end;

  If BeginsWith(Txt,'[Choice]') then begin
    Delete(Txt,1,8); Inc(X,55); Continue;
  end;

  I := Byte(Txt[1]);
  Delete(Txt,1,1);
  Case I of
    9:  Inc(X,25);
    13: begin Result.X := Max(Result.X,X); X := 15; Inc(Y,30); end;
    10: begin X := 15; Result.X := Max(Result.X,X); end;
    else Inc(X,Text_Sizes[I-32]-1);
  end;
end;
Result.Y := Y;
Result.X := Max(Result.X,X);
end;

function TextBlockSizeNeeded(Name,Txt:String;Speech:Boolean):TPoint;
  procedure AddS(S:String);
  var
  Ad:String;
  begin
  Ad := '';
  If EndsWith(S,#13#10) then begin Delete(S,Length(S)-1,2); Ad := #13#10; end;
  If Name<>'' then begin
    Result := MaxPoint(Result,TextSizeNeeded(Name+#13#10#147+S+#127+Ad)); Name := ''; Exit;
  end;
  If Speech then Result := MaxPoint(Result,TextSizeNeeded(#147+S+#127+Ad))
   else Result := MaxPoint(Result,TextSizeNeeded(S+Ad));
  end;
var
Ps:       Integer;
begin
Result := Point(0,0);
Repeat
  Ps := Pos(#13#10'[NewScreen]'#13#10,Txt);
  If Ps=0 then Break;
  AddS(Copy(Txt,1,Ps+1));
  Delete(Txt,1,Ps+14);
Until False;
If Txt<>'' then AddS(Txt);
end;

procedure InitArrays;
var
I,J:    Integer;
begin
Log('InitArrays');
 For I := $0 to $FF do chPC[I] := Char(NA);
 For I := $0 to $FF do
  For J := $0 to $FF do
    If FF7Ch[J]=I then chPC[I] := Char(J);

	chPC[$E2] := ' ';
	chPC[$EA] := ' ';
	chPC[$EB] := ' ';
	chPC[$EC] := ' ';
	chPC[$ED] := ' ';
	chPC[$EE] := ' ';
	chPC[$D5] := ' ';	// D5: Cyan (following FE)
	chPC[$D6] := ' ';	// D5: Green
	chPC[$D7] := ' ';	// D7: Purple (following FE)
	chPC[$D9] := ' ';	// D9: White (following FE)
	chPC[$E0] := ' ';	// E0
	chPC[$E1] := #9;	// E1: Tab
	chPC[$E7] := #10;	// E7: Line feed
	chPC[$E8] := #13;	// E8: Dialogue feed
	chPC[$F0] := ' ';	// F0
	chPC[$F2] := ' ';	// F0
	chPC[$FE] := ' ';	// FE: Change text color
	chPC[$FF] := #0;	// FF: End of string (NULL)

For I := 0 to 255 do
  StPC[I] := chPC[I];

StPC[$FE] := '[Colour:';
StPC[$D5] := 'Cyan]';
StPC[$D6] := 'Green]';
StPC[$D7] := 'Purple]';
StPC[$D9] := 'White]';
StPC[$DF] := '...]';
StPC[$DC] := 'grey?]';
StPc[$FF] := #13#10'[End of dialogue]'#13#10;
StPC[$A9] := '...';
StPC[$B2] := '[Begin]';
StPC[$B3] := '[End]';

StPC[$E0] := '[Choice]';
StPC[$E2] := ', ';
StPC[$E7] := #13#10;
StPC[$E8] := #13#10+'[NewScreen]'+#13#10;
StPC[$EA] := '[Cloud]';
StPC[$EB] := '[Barret]';
StPC[$EC] := '[Tifa]';
StPC[$ED] := '[Aeris]';
StPC[$EE] := '[Red XIII]';
StPC[$EF] := '[Yuffie]';
StPC[$F0] := '[Cait Sith]';
StPC[$F1] := '[Vincent]';
StPC[$F2] := '[Cid]';
StPC[$F3] := '[Party #1]';
StPC[$F4] := '[Party #2]';
StPC[$F5] := '[Party #3]';
StPC[$F6] := '[CIRCLE]';
StPC[$F7] := '[SQUARE]';
StPC[$F8] := '[TRIANGLE]';
StPC[$F9] := '[X]';

For I := 0 to $FF do
  If Length(StPC[I])>1 then InSt := InSt + [StPC[I][1]];

Log('InitArrays done');
end;

initialization
InitArrays;
end.
