unit FFLZs;

{Two versions of most functions. The versions ending xxxS are the stream
versions. Probably more safe/stable, but also probably slower.}

interface

Uses Classes, Windows, DebugLog;

Type
  PByte = ^Byte;
  TByte4k = Array[0..4115] of Byte;
  TByte18 = Array[0..17] of Byte;
  LZSCallback = procedure (Cur,Max:Integer);
  PByte4k = ^TByte4k;
  PByte18 = ^TByte18;
  TSeekProc = procedure (var Pre4k:TByte4k;var ThisData:TByte18;var SeqLen:Integer; var SeekBack:Integer);

  TUltraDodgySpec = record
    OriginalCompressed,NewData:         TMemoryStream;
    MatchLen,OldDiffLen,NewDiffLen:     Integer;
  end;

  TSteppedLZS = class(TObject)
    private
      FDone:       Boolean;
      procedure DoCompress;
      procedure DoDecompress;
    public
      DstLen:      Integer;
      Compressing: Boolean;
      Src,Dst:  TMemoryStream;
      property Done: Boolean read FDone;
      procedure Compress(Source:TMemoryStream);
      procedure Decompress(Source:TMemoryStream);
      procedure Step;
      Constructor Create;
      Destructor Destroy; Override;
  end;
  TQhimmCompress = function(Src:Pointer;SrcLen:Integer;Dst:Pointer;var DstLen:Integer): LongBool; cdecl;

Var
LZOFFSET:Integer = 18;
LZLEN: Integer=3;
CALLBACK: LZSCallback = nil;
LzPointerSeek: TSeekProc;
QhimmCompress: TQhimmCompress = nil;
QhimmDLL: THandle = 0;

function lzs_decompress(Src_First:Pointer;Src_Len:Integer;Dst_First:Pointer;var Dst_Len:Integer):Boolean;
function lzs_decompressS(Src:TMemoryStream):TMemoryStream;

procedure DodgySeek0(var Pre4k:TByte4k;var ThisData:TByte18;var SeqLen:Integer; var SeekBack:Integer);
procedure DodgySeek8(var Pre4k:TByte4k;var ThisData:TByte18;var SeqLen:Integer; var SeekBack:Integer);
procedure DodgySeek16(var Pre4k:TByte4k;var ThisData:TByte18;var SeqLen:Integer; var SeekBack:Integer);
procedure DodgySeek32(var Pre4k:TByte4k;var ThisData:TByte18;var SeqLen:Integer; var SeekBack:Integer);
procedure DodgySeekQhimm(var Pre4k:TByte4k;var ThisData:TByte18;var SeqLen:Integer; var SeekBack:Integer);

function LZSMemDecompress(Src:TMemoryStream):TMemoryStream;
function LZSFileDecompress(FName:String):TMemoryStream;
function LZSFileDecompressS(FName:String):TMemoryStream;
function LZSFileCompressS(FName:String):TMemoryStream;
function DodgyCompressS(Src:TStream): TMemoryStream;
function DodgyCompressSO(Src:TStream): TMemoryStream;
function _DodgyCompressS(Src:TStream): TMemoryStream;
//function lzs_compress(Src_First: Pointer; Src_Len:Integer; Dst_First: Pointer; var Dst_Len:Integer): Boolean;

function UltraDodgyCompress(Details:TUltraDodgySpec): TMemoryStream;

function LzsQhimmCompress(Src:TMemoryStream):TMemoryStream;

Const
  ITEMMAX = 16;
  FLAG_BYTES = 4;
  FLAG_COMPRESS = 0;
  FLAG_COPY = 1;

implementation

function Min(V1,V2:Integer):Integer;
begin
If V1<V2 then Result := V1 else Result := V2;
end;
function Max(V1,V2:Integer):Integer;
begin
If V1>V2 then Result := V1 else Result := V2;
end;

function LZSFileDecompress(FName:String):TMemoryStream;
var
Src: TMemoryStream;
begin
Src := TMemoryStream.Create;
Src.LoadFromFile(FName);
Result := LZSMemDecompress(Src);
Src.Free;
end;

function LZSFileDecompressS(FName:String):TMemoryStream;
var
Src: TMemoryStream;
begin
Src := TMemoryStream.Create;
Src.LoadFromFile(FName);
Result := LZS_DecompressS(Src);
Src.Free;
end;

function LZSFileCompressS(FName:String):TMemoryStream;
var
Src: TMemoryStream;
begin
Src := TMemoryStream.Create;
Src.LoadFromFile(FName);
Result := DodgyCompressSO(Src);
Src.Free;
end;

function LZSMemDecompress(Src:TMemoryStream):TMemoryStream;
var
I:      Integer;
begin
Result := TMemoryStream.Create;
Result.Size := Src.Size * 8;
lzs_decompress(Src.Memory,Src.Size,Result.Memory,I);
Result.Size := I;
end;

function PS(var P:PByte; var S:PByte):Boolean;
begin
Result := P^ = S^;
Inc(P); Inc(S);
end;
{
function lzs_compress(Src_First: Pointer; Src_Len:Integer; Dst_First: Pointer; var Dst_Len:Integer): Boolean;
var
Src1,Src2,Src,Dst,Src_Post,Dst_Post,Src_Max1,Src_Max16,P_Control: PByte;
Hash: Array[0..4095] of PByte;
P,S:  PByte;
Control, ControlBits:     Word;
Unroll,Len,Index:         Word;
Offset:                   Longint;
Value:                    Boolean;
PI:                       PInteger;
Label begin_unrolled_loop;
label literal;
Label end_unrolled_loop;
begin
Src := Src_First; Dst := Dst_First;
Src_Post := Pointer( Integer(Src_First) + Src_Len );
Dst_Post := Pointer( Integer(Dst_First) + Src_Len );
Src_Max1 := Pointer( Integer(Src_Post) - ITEMMAX );
Src_Max16 := Pointer( Integer(Src_Post) - 8*ITEMMAX );
Control := 0; ControlBits := 0;
Dst^ := FLAG_COMPRESS;
Inc(Dst,FLAG_BYTES);
P_Control := Dst;
Inc(Dst);

While True do begin
  Unroll := 8;
  If Integer(Src) > Integer(Src_Max16) then begin
    Unroll := 1;
    If (Integer(Src) > Integer(Src_Max1)) then begin
      If Src = Src_Post then Break;
      Goto Literal;
    end;
  end;

  begin_unrolled_loop:
  Src1 := Src; Inc(Src1); Src2 := Src; Inc(Src2,2);
  Index := ((40543 * ((((Src^ shl 4) xor Src1^) shl 4) xor Src2^)) shr 4) and $FFF;
  P := Hash[Index];
  S := Src;
  Hash[Index] := S;
  Offset := Integer(S) - Integer(P);
  Value := (Offset>4095) or (Integer(P)<Integer(Src_First)) or (Offset=0) or PS(P,S) or PS(P,S) or PS(P,S);

  If Value then begin
  literal:
    Dst^ := Src^; Inc(Dst); Inc(Src);
    Control := Control shr 1;
    Inc(ControlBits);
  end else begin
    PS(P,S); PS(P,S); PS(P,S); PS(P,S); PS(P,S); PS(P,S); PS(P,S);
    PS(P,S); PS(P,S); PS(P,S); PS(P,S); PS(P,S); PS(P,S);
    Inc(S);
    Len := Integer(S) - Integer(Src) - 1;
    Offset := ((Integer(Src)-Integer(Src_First))-Offset-18) and $FFF;
    Dst^ := Byte(Offset and $FF); Inc(Dst);
    Dst^ := Byte( ((Offset and $F00) shr 4) + (Len-3)); Inc(Dst);
    Inc(Src,Len);
    Control := (Control shr 1) or $80;
    Inc(ControlBits);
  end;
end_unrolled_loop:

    Dec(Unroll);
    If (Unroll<>0) then goto begin_unrolled_loop;

		if (controlbits=8) then begin
      P_Control^ := Control xor $FF;
      P_Control := Dst;
      Inc(Dst);
      ControlBits := 0;
      Control := ControlBits;
    end;
end;

Control := Control shr (8-ControlBits);
P_Control^ := Control xor $FF;
Inc(P_Control);
if (P_Control = Dst) then Dec(Dst);
Dst_Len := Integer(Dst) - Integer(Dst_First);
PI := Dst_First;
PI^ := Dst_Len - 4;
end;
}

function lzs_decompress(Src_First:Pointer;Src_Len:Integer;Dst_First:Pointer;var Dst_Len:Integer):Boolean;
var
ControlBits:          Word;
Control:              Byte;
Src,Dst,Post:         PByte;
Offset,Len:           Word;
P:                    PByte;
begin
Result := False;
ControlBits := 0; Control := $FF;
Src := Src_First;
Dst := Dst_First;
Post := Pointer( Integer(Src_First) + Src_Len );
Inc(Src,4);
while Src <> Post do begin
  If ControlBits=0 then begin
    Control := Src^;
    Inc(Src);
    ControlBits := 8;
  end;
  If (Control and $1)=0 then begin
    Offset := Src^ and $FF;
    Inc(Src);
    Inc(Offset, (Src^ and $F0) shl 4);
    Len := LZLEN + (Src^ and $F);
    Inc(Src);

    P := Pointer( Integer(Dst) - (   (   (Integer(Dst)-Integer(Dst_First)) - (Offset+LZOFFSET)) and $FFF));

    While Len>0 do begin
      Dec(Len);
      If Integer(P) < Integer(Dst_First) then begin
        Dst^ := 0; Inc(Dst);
        Inc(P);
      end else begin
        Dst^ := P^; Inc(Dst); Inc(P);
      end;
    end;
  end else begin
    Dst^ := Src^; Inc(Dst); Inc(Src);
  end;
  Control := Control shr 1;
  Dec(ControlBits);
end;
Dst_Len := Integer(Dst) - Integer(Dst_First);
Result := True;
end;

function lzs_decompressS(Src:TMemoryStream):TMemoryStream;
var
ControlBits:          Word;
Control:              Byte;
Offset,Len:           Word;
I,SeekAmt,DstLen,BTmp:  Integer;
Other,Buffer:         Array[0..63] of Byte; //larger for safety. Hey.
begin
Result := TMemoryStream.Create;
Result.Size := Src.Size * 8;
ControlBits := 0; Control := $FF;
DstLen := 0;
Src.Position := 4;
Log('Setup to decompress');
while Src.Position < Src.Size do begin
  If @Callback <> nil then
    If (Src.Position mod 1024)=0 then
      Callback(Src.Position,Src.Size);
  If ControlBits=0 then begin
    Src.ReadBuffer(Control,1);
    ControlBits := 8;
  end;
  If (Control and $1)=0 then begin
    Src.ReadBuffer(BTmp,1);
    Offset := BTmp and $FF;
    Src.ReadBuffer(BTmp,1);
    Inc(Offset, (BTmp and $F0) shl 4);
    Len := LZLEN + (BTmp and $F);
    SeekAmt := (   (   (Result.Position) - (Offset+LZOFFSET)) and $FFF);
    If SeekAmt <= Result.Position then begin
      Result.Seek( -SeekAmt, soFromCurrent);

      Result.ReadBuffer(Buffer,Len);

      If Len > SeekAmt then begin //dodgy hack moment!
        For I := 0 to (Len div SeekAmt)+1 do
          CopyMemory(@Other[I*SeekAmt],@Buffer,SeekAmt);
        CopyMemory(@Buffer,@Other,24);
      end;


      Result.Position := DstLen;
    end else ZeroMemory(@Buffer,Len);
    Result.WriteBuffer(Buffer,Len);
    Inc(DstLen,Len);

  end else begin
    Src.ReadBuffer(BTmp,1);
    Result.WriteBuffer(BTmp,1);
    Inc(DstLen);
  end;
  Control := Control shr 1;
  Dec(ControlBits);
end;
Log('Done!');
  If @Callback <> nil then
    Callback(Src.Size,Src.Size);
Result.Size := DstLen;
end;

procedure DodgySeek32(var Pre4k:TByte4k;var ThisData:TByte18;var SeqLen:Integer; var SeekBack:Integer);
var
MaxSB,MaxSF,DataPos,I:    Integer;
PI1,PI2:                            PInteger;
PB1,PB2:                            PByte;
begin
MaxSF := 0; MaxSB := $FFF - SeekBack;
For I := 4093 downto MaxSB do begin
  DataPos := 0;
  PI1 := @Pre4k[I];
  PI2 := @ThisData;
  While (PI1^=PI2^) and (DataPos < 18) and (I+DataPos<4087) do begin
    Inc(DataPos,4); Inc(PI1); Inc(PI2);
  end;
  If DataPos>17 then DataPos := 18 else
    If (DataPos<>0) then begin
      PB1 := Pointer(PI1); PB2 := Pointer(PI2);
      While (PB1^=PB2^) and (DataPos < 18) do begin
        Inc(PB1); Inc(PB2); Inc(DataPos);
      end;
    end;

{  For J := I to Min(I+17,4095) do
    If Pre4k[J] = ThisData[DataPos] then Inc(DataPos) else Break;}
  If (DataPos > MaxSF)then begin
    MaxSF := DataPos;
    SeekBack := $FFF - I;
    If (DataPos=18) then Break;
  end;
end;
If MaxSF>2 then SeqLen := MaxSF else SeqLen := 0;
end;

procedure DodgySeek16(var Pre4k:TByte4k;var ThisData:TByte18;var SeqLen:Integer; var SeekBack:Integer);
var
MaxSB,MaxSF,DataPos,I:    Integer;
PI1,PI2:                            PWord;
PB1,PB2:                            PByte;
begin
MaxSF := 0; MaxSB := $FFF - SeekBack;
For I := 4093 downto MaxSB do begin
  DataPos := 0;
  PI1 := @Pre4k[I];
  PI2 := @ThisData;
  While (PI1^=PI2^) and (DataPos < 18) and (I+DataPos<4093) do begin
    Inc(DataPos,2); Inc(PI1); Inc(PI2);
  end;
  If (DataPos<>18) and (DataPos<>0) then begin
    PB1 := Pointer(PI1); PB2 := Pointer(PI2);
    If PB1^ = PB2^ then Inc(DataPos);
  end;

{  For J := I to Min(I+17,4095) do
    If Pre4k[J] = ThisData[DataPos] then Inc(DataPos) else Break;}
  If (DataPos > MaxSF)then begin
    MaxSF := DataPos;
    SeekBack := $FFF - I;
    If (DataPos=18) then Break;
  end;
end;
If MaxSF>2 then SeqLen := MaxSF else SeqLen := 0;
end;

procedure DodgySeek0(var Pre4k:TByte4k;var ThisData:TByte18;var SeqLen:Integer; var SeekBack:Integer);
begin
SeqLen := 0;
end;

procedure DodgySeek8(var Pre4k:TByte4k;var ThisData:TByte18;var SeqLen:Integer; var SeekBack:Integer);
var
MaxSB,MaxSF,DataPos,I,J:    Integer;
begin
MaxSF := 0; MaxSB := $FFF - SeekBack;
For I := 4093 downto MaxSB do begin
  DataPos := 0;
  For J := I to Min(I+17,4093) do
    If Pre4k[J] = ThisData[DataPos] then Inc(DataPos) else Break;
  If (DataPos > MaxSF)then begin
    MaxSF := DataPos;
    SeekBack := $FFF - I;
    If (DataPos=18) then Break;
  end;
end;
If MaxSF>2 then SeqLen := MaxSF else SeqLen := 0;
end;

function _DodgyCompressS(Src:TStream): TMemoryStream;
var
Len: Integer;
B:  Byte;
Buf:Array[0..7] of Byte;
begin
Len := Src.Size;
Len := Len + (Len div 8);
If (Src.Size mod 8) <> 0 then Inc(Len);
Result := TMemoryStream.Create;
Result.Size := Len+4;
Result.Position := 0;
Result.WriteBuffer(Len,4);
B := $FF;
While Src.Position < Src.Size do begin
  Len := Min(8,Src.Size-Src.Position);
  Result.WriteBuffer(B,1);
  Src.ReadBuffer(Buf,Len);
  Result.WriteBuffer(Buf,Len);
end;
end;

function DodgyCompressS(Src:TStream): TMemoryStream;
var
Control,ControlBits,BTmp: Byte;
NumNull,SeqLen,StPos,Tmp,SeekAmt,TmpOffset,DstLen,LastControl: Integer;
Offset:                   Word;
Last4k:                   TByte4k;
ThisData:                 TByte18;
NV2,NV:                   Smallint;
PB:                       PByte;
begin
Result := TMemoryStream.Create;
Result.Size := Src.Size;
Result.Position := 5; LastControl := 4;
DstLen := 4;
Control := 0; ControlBits := 0;
While Src.Position < Src.Size do begin
  StPos := Src.Position;
  If (StPos mod 1024)=0 then
    If @Callback <> nil then
      Callback(StPos,Src.Size);
//  SeekAmt := Min(StPos,$FFF);
  SeekAmt := $FFF;
  ZeroMemory(@Last4k,Sizeof(Last4k));
  Src.Position := Max(0,Src.Position-$FFF);
  PB := @Last4k[$FFF-Min(StPos,$FFF)];
  Src.ReadBuffer(PB^,Min(StPos,$FFF));
  Src.ReadBuffer(ThisData,Min(18,Src.Size-Src.Position));
//  Src.Position := Tmp+1;
  DodgySeek32(Last4k,ThisData,SeqLen,SeekAmt);
//  If SeekAmt > StPos then SeekAmt := StPos + 1;
  If SeekAmt > StPos then begin
    NumNull := 0;
    While PB^=0 do begin
      Inc(NumNull); Inc(PB);
    end;
    SeekAmt := StPos+1;
    SeekAmt := SeekAmt + (SeqLen - 1 - NumNull);
  end;
//  If SeekAmt > StPos then SeqLen := 0;
//  SeqLen := Min(SeqLen,Src.Size-Src.Position);
  If SeqLen<3 then begin
     SeqLen := 1;
     Control := Control or (1 shl ControlBits);
     Result.WriteBuffer(ThisData[0],1);
     Inc(DstLen);
  end else begin
    NV := SeekAmt;
    {If NV > (Src.Position-LZOFFSET) then }NV := NV or ($F000);
    Offset := StPos - LZOFFSET - NV;

    BTmp := (Offset and $FF);
    Result.WriteBuffer(BTmp,1);
    BTmp := ((Offset and $F00) shr 4) or ((SeqLen-3) and $F);
    Result.WriteBuffer(BTmp,1);
    Inc(DstLen,2);
  end;
  Src.Position := StPos + SeqLen;
  Inc(ControlBits);
  If (ControlBits=8) or (Src.Position >= Src.Size) then begin
    Tmp := Result.Position;
    Result.Position := LastControl;
    Result.WriteBuffer(Control,1);
    Inc(DstLen);
    Result.Position := Tmp+1;
    LastControl := Tmp;
    Control := 0; ControlBits := 0;
  end;
end;
Result.Size := DstLen;
Result.Position := 0;
Tmp := DstLen - 4;
Result.WriteBuffer(Tmp,4);
end;

function DodgyCompressSO(Src:TStream): TMemoryStream;
var
Control,ControlBits: Byte;
SeqLen,Tmp,SeekAmt,DstLen: Integer;
Offset:                   Word;
NV:                       Smallint;
LastCtrl,Dst,Dest,Last4k,ThisData,Cur,Data,PB:              PByte;
ICur:                     Integer;
begin
If (@LzPointerSeek = @DodgySeekQhimm) and (Src is TMemoryStream) then begin
  Result := LzsQhimmCompress(TMemoryStream(Src));
  Exit;
end;

DstLen := Src.Size + (Src.Size div 8) + 1;
GetMem(Dest,DstLen);
Dst := Dest;
DstLen := 4;
LastCtrl := Dest;
Inc(Dst);

Control := 0; ControlBits := 0;
GetMem(Data,Src.Size+4095+64);
ZeroMemory(Data,Src.Size+4095+64);
Cur := Data; Inc(Cur,4095);
Src.Position := 0;
Src.ReadBuffer(Cur^,Src.Size);
Last4k := Pointer(Data);
ThisData := Pointer(Cur);
Inc(Cur,Src.Size);
ICur := Integer(Cur);
While Integer(ThisData) < ICur do begin
  If (Integer(ThisData) mod 1024)=0 then
    If @Callback <> nil then
      Callback(Integer(ThisData)-Integer(Data),Src.Size);
//  SeekAmt := Min(Integer(ThisData)-Integer(Data),$FFF);
  SeekAmt := $FFF;
  try
  LzPointerSeek(PByte4k(Last4k)^,PByte18(ThisData)^,SeqLen,SeekAmt);
  except
  end;

  If SeqLen<3 then begin
     SeqLen := 1;
     Control := Control or (1 shl ControlBits);
     Dst^ := PByte18(ThisData)^[0]; Inc(Dst);
     Inc(DstLen);
  end else begin
    NV := SeekAmt;
    NV := NV or ($F000);
    Offset := (Integer(ThisData) - Integer(Data) - 4095) - LZOFFSET - NV;

    Dst^ := (Offset and $FF); Inc(Dst);
    Dst^ := ((Offset and $F00) shr 4) or ((SeqLen-3) and $F);
    Inc(Dst);
    Inc(DstLen,2);
  end;

  Inc(Last4k,SeqLen); Inc(ThisData,SeqLen);

  Inc(ControlBits);
  If (ControlBits=8) or (Integer(ThisData) >= ICur) then begin
    LastCtrl^ := Control;
    LastCtrl := Dst;
    Inc(Dst);
    Inc(DstLen);
    Control := 0; ControlBits := 0;
  end;
end;
Result := TMemoryStream.Create;
Result.Size := DstLen;
Result.Position := 0;
Tmp := DstLen - 4;
Result.WriteBuffer(Tmp,4);
Result.WriteBuffer(Dest^,DstLen-4);
FreeMem(Dest);
FreeMem(Data);
If @Callback <> nil then
  Callback(Src.Size,Src.Size);
end;

procedure TSteppedLZS.Compress(Source:TMemoryStream);
var
I:Integer;
begin
Src.Clear;
Src.Position := 0;
Src.CopyFrom(Source,0);
Src.Position := 0;
Compressing := True;
Dst.Clear;
I := 0;
Dst.WriteBuffer(I,4);
DstLen := 4;
FDone := False;
end;

procedure TSteppedLZS.Decompress(Source:TMemoryStream);
begin
Src.Clear;
Src.Position := 0;
Src.CopyFrom(Source,0);
Src.Position := 4;
Compressing := False;
Dst.Clear;
Dst.Size := Src.Size * 9;
DstLen := 0;
FDone := False;
end;

procedure TSteppedLZS.Step;
begin
FDone := False;
If Compressing then DoCompress else DoDecompress;
end;

Constructor TSteppedLZS.Create;
begin
Inherited;
Src := TMemoryStream.Create;
Dst := TMemoryStream.Create;
end;

Destructor TSteppedLZS.Destroy;
begin
Src.Free; Dst.Free;
Inherited;
end;

procedure TSteppedLZS.DoCompress;
var
Control,ControlBits,BTmp: Byte;
NumNull,SeqLen,StPos,Tmp,SeekAmt,TmpOffset,LastControl: Integer;
Offset:                   Word;
Last4k:                   TByte4k;
ThisData:                 TByte18;
NV2,NV:                   Smallint;
PB:                       PByte;
begin
LastControl := Dst.Position;
Dst.Position := LastControl+1;
Control := 0; ControlBits := 0;
While (Src.Position < Src.Size) and (ControlBits<8) do begin
  StPos := Src.Position;
  SeekAmt := $FFF;
  ZeroMemory(@Last4k,Sizeof(Last4k));
  Src.Position := Max(0,Src.Position-$FFF);
  PB := @Last4k[$FFF-Min(StPos,$FFF)];
  Src.ReadBuffer(PB^,Min(StPos,$FFF));
  Src.ReadBuffer(ThisData,Min(18,Src.Size-Src.Position));
  LzPointerSeek(Last4k,ThisData,SeqLen,SeekAmt);
//  If SeqLen>SeekAmt then SeqLen := SeekAmt;
  If SeekAmt > StPos then begin
    NumNull := 0;
    While PB^=0 do begin
      Inc(NumNull); Inc(PB);
    end;
    SeekAmt := StPos+1;
    SeekAmt := SeekAmt + (SeqLen - 1 - NumNull);
  end;
  If (SeqLen<3) then begin
     SeqLen := 1;
     Control := Control or (1 shl ControlBits);
     Dst.WriteBuffer(ThisData[0],1);
     Inc(DstLen);
  end else begin
    NV := SeekAmt;
    NV := NV or ($F000);
    Offset := StPos - LZOFFSET - NV;

    BTmp := (Offset and $FF);
    Dst.WriteBuffer(BTmp,1);
    BTmp := ((Offset and $F00) shr 4) or ((SeqLen-3) and $F);
    Dst.WriteBuffer(BTmp,1);
    Inc(DstLen,2);
  end;
  Src.Position := StPos + SeqLen;
  Inc(ControlBits);
end;
Tmp := Dst.Position;
Dst.Position := LastControl;
Dst.WriteBuffer(Control,1);
Inc(DstLen);
Dst.Position := Tmp;
FDone := Src.Position >= Src.Size;
If FDone then begin
  Dst.Size := DstLen;
  Dst.Position := 0;
  Tmp := DstLen - 4;
  Dst.WriteBuffer(Tmp,4);
end;
end;

procedure TSteppedLZS.DoDecompress;
var
ControlBits:          Word;
Control:              Byte;
Offset,Len:           Word;
I,SeekAmt,BTmp:  Integer;
Other,Buffer:         Array[0..63] of Byte; //larger for safety. Hey.
begin
ControlBits := 8;
Src.ReadBuffer(Control,1);
While (Src.Position < Src.Size) and (ControlBits > 0) do begin
  If (Control and $1)=0 then begin
    Src.ReadBuffer(BTmp,1);
    Offset := BTmp and $FF;
    Src.ReadBuffer(BTmp,1);
    Inc(Offset, (BTmp and $F0) shl 4);
    Len := LZLEN + (BTmp and $F);
    SeekAmt := (   (   (Dst.Position) - (Offset+LZOFFSET)) and $FFF);
    If SeekAmt <= Dst.Position then begin
      Dst.Seek( -SeekAmt, soFromCurrent);

      Dst.ReadBuffer(Buffer,Len);

      If Len > SeekAmt then begin //dodgy hack moment!
        For I := 0 to (Len div SeekAmt)+1 do
          CopyMemory(@Other[I*SeekAmt],@Buffer,SeekAmt);
        CopyMemory(@Buffer,@Other,24);
      end;


      Dst.Position := DstLen;
    end else ZeroMemory(@Buffer,Len);
    Dst.WriteBuffer(Buffer,Len);
    Inc(DstLen,Len);

  end else begin
    Src.ReadBuffer(BTmp,1);
    Dst.WriteBuffer(BTmp,1);
    Inc(DstLen);
  end;
  Control := Control shr 1;
  Dec(ControlBits);
end;
FDone := Src.Position >= Src.Size;
If FDone then Dst.Size := DstLen;
end;

function UltraDodgyCompress(Details:TUltraDodgySpec): TMemoryStream;
var
Fooled,Stepper:      TSteppedLZS;
Change,I:            Integer;
begin
Stepper := TSteppedLZS.Create;
Stepper.Decompress(Details.OriginalCompressed);

Change := Details.NewDiffLen - Details.OldDiffLen;

Result := TMemoryStream.Create;
I := 0;
Result.Size := Details.NewData.Size;
Result.WriteBuffer(I,4);
While Stepper.Dst.Position < (Details.MatchLen - 24) do
  Stepper.Step;
I := Stepper.Src.Position;
Stepper.Src.Position := 4;
Result.CopyFrom(Stepper.Src,I-4);
Fooled := TSteppedLZS.Create;
Fooled.Src.Free; Fooled.Dst.Free;
Fooled.Src := Details.NewData;
Fooled.Dst := Result;
Details.NewData.Position := Stepper.Dst.Position;
Fooled.Compressing := True;
Fooled.DstLen := I;
While Fooled.Src.Position < (Details.MatchLen + Details.NewDiffLen) do
  Fooled.Step;
While (Not Fooled.Done) and (Fooled.Src.Position <> (Stepper.Dst.Position+Change)) do begin
  While (Stepper.Dst.Position+Change) < Fooled.Src.Position do Stepper.Step;
  While Fooled.Src.Position < (Stepper.Dst.Position+Change) do Fooled.Step;
end;
If Not Fooled.Done then begin
  Fooled.Dst.Size := Fooled.DstLen + (Stepper.Src.Size - Stepper.Src.Position);
  Fooled.Dst.CopyFrom(Stepper.Src,Stepper.Src.Size - Stepper.Src.Position);
end;
Fooled.Dst := TMemoryStream.Create;
Fooled.Src := TMemoryStream.Create;
Fooled.Free; Stepper.Free;
I := Result.Size-4;
Result.Position := 0;
Result.WriteBuffer(I,4);
I := Result.Size;

end;

function LzsQhimmCompress(Src:TMemoryStream):TMemoryStream;
var
Len:  Integer;
begin
If @QhimmCompress=nil then begin
  QhimmDLL := LoadLibrary('QHIMMLZS.DLL');
  If QhimmDLL=0 then begin
    LzPointerSeek := @DodgySeek16;
    Result := DodgyCompressSO(Src);
    Exit;
  end;
  @QhimmCompress := GetProcAddress(QhimmDLL,'_lzs_compress');
end;
Result := TMemoryStream.Create;
Result.Size := Src.Size + (Src.Size div 8) + 8;
Len := Result.Size;
QhimmCompress(Src.Memory,Src.Size,Result.Memory,Len);
Result.Size := Len;
end;

procedure DodgySeekQhimm(var Pre4k:TByte4k;var ThisData:TByte18;var SeqLen:Integer; var SeekBack:Integer);
begin
end;

initialization
LzPointerSeek := DodgySeek16;
finalization
If QhimmDLL<>0 then FreeLibrary(QhimmDLL);
end.
