unit kapgrad;

interface

{$DEFINE APIPIXELS}
//Define APIPIXELS to use Get/Set Pixel API calls rather than
//using Pixels[]. No reason to use Pixels[] atall, except for
//speed comparisons (gen. at least 3x slower)

{$DEFINE BUFFERED_4_GRADIENT}

uses Windows,Classes,SysUtils,Graphics;

Type
    TFourLong = array[0..3] of Longint;
    TFourColours = record
                Case Integer of
                0: (Dat:TFourLong);
                1: (TL,BL,TR,BR:TColor);
    end;

Const
     ffColours : TFourColours = (Dat: ($B05800,$800000,$500000,$200000) );
     capColours: TFourColours = (Dat: ($808000,$000000,$FFFFFF,$808000) );

procedure FillGrad(St,En:TColor;Width,Height:Integer;Can:TCanvas);
procedure SquareGrad(Can:TCanvas;Buffered:Boolean;Width,Height:Integer;
          Colours:TFourColours);
function FourColours(TL,BL,TR,BR:Longint):TFourColours;
procedure ffBorder(Can:TCanvas;Width,Height,Border:Integer);
procedure Fil(Can:TCanvas;Width,Height:Integer;Colours:TFourColours);
procedure FilOpt(Bmp:TBitmap;Width,Height:Integer;Colours:TFourColours);
function Grey(Val:Byte):Integer;
function FourColsSame(F1,F2:TFourColours):Boolean;
procedure FilCentreColour(Can:TCanvas;Width,Height:Integer;Colours:TFourColours;Centre:TColor);

implementation

{$IFDEF BUFFERED_4_GRADIENT}
var
   Buffer:    TBitmap;          //Stores a buffer of recently-drawn gradient
   BufCols:   TFourColours;     //If a smaller gradient w.same colours needs
              //to be drawn, the buffer can be used
{$ENDIF}
function FourColours(TL,BL,TR,BR:Longint):TFourColours;
var FC:  TFourColours;
begin
FC.TL := TL; FC.BL := BL;
FC.TR := TR; FC.BR := BR;
Result := FC;
end;

procedure FillGrad(St,En:TColor;Width,Height:Integer;Can:TCanvas);
var
R,G,B,I:     Integer;
dR,dG,dB:    Real;
begin
R := St and $FF;
G := (St shr 8) and $FF;
B := (St shr 16) and $FF;
dR := (En and $FF) - R;
dG := ((En shr 8) and $FF) - G;
dB := ((En shr 16) and $FF) - B;
For I := 0 to Width do begin
    Can.MoveTo(I,0);
    Can.Pen.Color := (R + Trunc(dR * (I/Width))) or
                     ( (G + Trunc(dG * (I/Width))) shl 8) or
                     ( (B + Trunc(dB * (I/Width))) shl 16);
    Can.LineTo(I,Height);
end;

end;

procedure Fil(Can:TCanvas;Width,Height:Integer;Colours:TFourColours);
var
Tmp,R,G,B,dR,dG,dB,I,J:      Integer;
begin
{$IFDEF BUFFERED_4_GRADIENT}
If (FourColsSame(Colours,BufCols)) and (Width <= Buffer.Width) and (Height <= Buffer.Width) then
   Can.StretchDraw(Rect(0,0,Width,Height),Buffer)
else begin
{$ENDIF}
For I := 0 to Width do begin
    R := (Colours.TL and $FF) + MulDiv(I,((Colours.TR and $FF)-(Colours.TL and $FF)),Width);
    dR := R -( (Colours.BL and $FF) + MulDiv(I,((Colours.BR and $FF)-(Colours.BL and $FF)),Width));
    G := ((Colours.TL shr 8) and $FF) + MulDiv(I,(((Colours.TR shr 8) and $FF)-((Colours.TL shr 8) and $FF)),Width);
    dG := G -( ((Colours.BL shr 8) and $FF) + MulDiv(I,(((Colours.BR shr 8) and $FF)-((Colours.BL shr 8) and $FF)),Width));
    B := ((Colours.TL shr 16) and $FF) + MulDiv(I,(((Colours.TR shr 16) and $FF)-((Colours.TL shr 16) and $FF)),Width);
    dB := B -( ((Colours.BL shr 16) and $FF) + MulDiv(I,(((Colours.BR shr 16) and $FF)-((Colours.BL shr 16) and $FF)),Width));
    For J := 0 to Height do begin
        Tmp := ( ( R - MulDiv(J,dR,Height) ) and $FF ) or
               ( ( ( G - MulDiv(J,dG,Height) ) and $FF) shl 8 ) or
               ( ( ( B - MulDiv(J,dB,Height) ) and $FF) shl 16 );
        SetPixel(Can.Handle,I,J,Tmp);
    end;
end;
{$IFDEF BUFFERED_4_GRADIENT}
If (Width - Buffer.Width) - (Height - Buffer.Height) > 0 then begin
   BufCols := Colours;
   Buffer.Width := Width; Buffer.Height := Height;
   Buffer.Canvas.CopyRect(Rect(0,0,Width,Height),Can,Rect(0,0,Width,Height));
end;
end;
{$ENDIF}
end;




procedure ChangePixel(P:PChar;X,Y,Width,Val:Integer);
var
PI:     ^Integer;
begin
PI := Pointer(P);
Inc(PI,Y*Width+X);
PI^ := Val;
end;

procedure FilOpt(Bmp:TBitmap;Width,Height:Integer;Colours:TFourColours);
var
Tmp,R,G,B,dR,dG,dB,I,J:      Integer;
PI1:                         ^Integer;
begin
{$IFDEF BUFFERED_4_GRADIENT}
If (FourColsSame(Colours,BufCols)) and (Width <= Buffer.Width) and (Height <= Buffer.Width) then
   Bmp.Canvas.StretchDraw(Rect(0,0,Width,Height),Buffer)
else begin
{$ENDIF}
{GetMem(Data,Width*Height*4);
ZeroMemory(Data,Width*Height*4);}
For I := 1 to Width do begin
    R := (Colours.TL and $FF) + MulDiv(I,((Colours.TR and $FF)-(Colours.TL and $FF)),Width);
    dR := R -( (Colours.BL and $FF) + MulDiv(I,((Colours.BR and $FF)-(Colours.BL and $FF)),Width));
    G := ((Colours.TL shr 8) and $FF) + MulDiv(I,(((Colours.TR shr 8) and $FF)-((Colours.TL shr 8) and $FF)),Width);
    dG := G -( ((Colours.BL shr 8) and $FF) + MulDiv(I,(((Colours.BR shr 8) and $FF)-((Colours.BL shr 8) and $FF)),Width));
    B := ((Colours.TL shr 16) and $FF) + MulDiv(I,(((Colours.TR shr 16) and $FF)-((Colours.TL shr 16) and $FF)),Width);
    dB := B -( ((Colours.BL shr 16) and $FF) + MulDiv(I,(((Colours.BR shr 16) and $FF)-((Colours.BL shr 16) and $FF)),Width));
    For J := 1 to Height do begin
        Tmp := ( ( R - MulDiv(J-1,dR,Height-1) ) and $FF ) or
               ( ( ( G - MulDiv(J-1,dG,Height-1) ) and $FF) shl 8 ) or
               ( ( ( B - MulDiv(J-1,dB,Height-1) ) and $FF) shl 16 ) or
               $FF000000;
        PI1 := Bmp.ScanLine[J-1];
        Inc(PI1,I-1);
        Tmp := (Tmp and $FF00FF00) or ( (Tmp and $FF) shl 16) or ((Tmp and $FF0000) shr 16);
        PI1^ := Tmp;
//        ChangePixel(Data,I-1,J-1,Width,Tmp);
    end;
end;
{For I := 0 to Height-1 do begin
  PI1 := Bmp.ScanLine[I];
  PC := Data;
  Inc(PC,Width*I*4);
  PI2 := Pointer(PC);
  CopyMemory(PI1,PI2,Width*4);
end;
FreeMem(Data);}
{$IFDEF BUFFERED_4_GRADIENT}
If (Width - Buffer.Width) - (Height - Buffer.Height) > 0 then begin
   BufCols := Colours;
   Buffer.Width := Width; Buffer.Height := Height;
   Buffer.Canvas.CopyRect(Rect(0,0,Width,Height),Bmp.Canvas,Rect(0,0,Width,Height));
end;
end;
{$ENDIF}
end;

procedure FilCentreColour(Can:TCanvas;Width,Height:Integer;Colours:TFourColours;Centre:TColor);
var
Tmp:      TFourColours;
Buf:      TBitmap;
MidL,R,G,B:    Integer;
begin
Buf := TBitmap.Create;
Buf.Width := Width div 2;
Buf.Height := Height div 2;
Tmp := Colours;
//1 - topleft square
    R := (Colours.TL and $FF) - ( ((Colours.TL and $FF)-(Colours.TR and $FF)) div 2);
    G := ((Colours.TL shr 8)and $FF) - (  (((Colours.TL shr 8)and $FF)-((Colours.TR shr 8)and $FF)) div 2);
    B := ((Colours.TL shr 16)and $FF) - ( (((Colours.TL shr 16)and $FF)-((Colours.TR shr 16)and $FF)) div 2);
    Tmp.TR := R or (G shl 8) or (B shl 16);
    R := (Colours.TL and $FF) - ( ((Colours.TL and $FF)-(Colours.BL and $FF)) div 2);
    G := ((Colours.TL shr 8)and $FF) - (  (((Colours.TL shr 8)and $FF)-((Colours.BL shr 8)and $FF)) div 2);
    B := ((Colours.TL shr 16)and $FF) - ( (((Colours.TL shr 16)and $FF)-((Colours.BL shr 16)and $FF)) div 2);
    Tmp.BL := R or (G shl 8) or (B shl 16);
    Tmp.BR := Centre;
    Fil(Buf.Canvas,Buf.Width,Buf.Height,Tmp);
    Can.Draw(0,0,Buf);
//2 - topright square
    MidL := Tmp.BL; //store for later
    Tmp.TL := Tmp.TR;
    Tmp.BL := Centre;
    Tmp.TR := Colours.TR;
    R := (Colours.TR and $FF) - ( ((Colours.TR and $FF)-(Colours.BR and $FF)) div 2);
    G := ((Colours.TR shr 8)and $FF) - (  (((Colours.TR shr 8) and $FF)-((Colours.BR shr 8)and $FF)) div 2);
    B := ((Colours.TR shr 16)and $FF) - ( (((Colours.TR shr 16)and $FF)-((Colours.BR shr 16)and $FF)) div 2);
    Tmp.BR := R or (G shl 8) or (B shl 16);
    Fil(Buf.Canvas,Buf.Width,Buf.Height,Tmp);
    Can.Draw(Buf.Width,0,Buf);
//3 - botleft square
    Tmp.TL := MidL;
    MidL := Tmp.BR; //now it's midR
    Tmp.TR := Centre;
    Tmp.BL := Colours.BL;
    R := (Colours.BL and $FF) - ( ((Colours.BL and $FF)-(Colours.BR and $FF)) div 2);
    G := ((Colours.BL shr 8)and $FF) - (  (((Colours.BL shr 8)and $FF)-((Colours.BR shr 8)and $FF)) div 2);
    B := ((Colours.BL shr 16)and $FF) - ( (((Colours.BL shr 16)and $FF)-((Colours.BR shr 16)and $FF)) div 2);
    Tmp.BR := R or (G shl 8) or (B shl 16);
    Fil(Buf.Canvas,Buf.Width,Buf.Height,Tmp);
    Can.Draw(0,Buf.Height,Buf);
//4 - botright square
    Tmp.BL := Tmp.BR;
    Tmp.TL := Centre;
    Tmp.TR := MidL;
    Tmp.BR := Colours.BR;
    Fil(Buf.Canvas,Buf.Width,Buf.Height,Tmp);
    Can.Draw(Buf.Width,Buf.Height,Buf);
Buf.Free;
end;


procedure SquareGrad(Can:TCanvas;Buffered:Boolean;Width,Height:Integer;
          Colours:TFourColours);
var
Tmp,Buf:  TBitmap;
Tgt:      TCanvas;
I,Max:    Integer;
begin
Tmp := TBitmap.Create;
If Width > Height then Max := Width else Max := Height;
Tmp.Height := 1; Tmp.Width := Max;
If Buffered then begin
   Buf := TBitmap.Create;
   Buf.Height := Height; Buf.Width := Width;
   Tgt := Buf.Canvas; end
else Tgt := Can;
FillGrad(Colours.TL,Colours.BL,Height,1,Tmp.Canvas);
For I := 0 to Height-1 do
{$IFDEF APIPIXELS}
    SetPixel(Tgt.Handle,0,I,GetPixel(Tmp.Canvas.Handle,I,0));
{$ELSE}
    Tgt.Pixels[0,I] := Tmp.Canvas.Pixels[I,0];
{$ENDIF}
FillGrad(Colours.TR,Colours.BR,Height,1,Tmp.Canvas);
For I := 0 to Height-1 do
{$IFDEF APIPIXELS}
    SetPixel(Tgt.Handle,Width-1,I,GetPixel(Tmp.Canvas.Handle,I,0));
{$ELSE}
    Tgt.Pixels[Width-1,I] := Tmp.Canvas.Pixels[I,0];
{$ENDIF}
For I := 0 to Height-1 do begin
    FillGrad(Tgt.Pixels[0,I],Tgt.Pixels[Width-1,I],Width-1,1,Tmp.Canvas);
    Tgt.Draw(0,I,Tmp);
end;
If Buffered then begin
   Can.Draw(0,0,Buf); Buf.Free;
end;
Tmp.Free;
end;

function Grey(Val:Byte):Integer;
begin
Result := Val or (Val shl 8) or (Val shl 16);
end;

procedure ffBorder(Can:TCanvas;Width,Height,Border:Integer);
begin
Can.Pen.Color := Grey(151); Can.Pen.Width := 2; Can.Brush.Style := bsClear;
Can.MoveTo(2,0); Can.LineTo(Width-5,0);
Can.Pen.Color := Grey(224); Can.MoveTo(4,2); Can.LineTo(Width-5,2);//Top
Can.LineTo(Width-3,4); Can.MoveTo(4,2); Can.LineTo(2,4);
Can.Pen.Color := Grey(79); Can.MoveTo(6,4); Can.LineTo(Width-7,4);
Can.MoveTo(6,4); Can.LineTo(4,6);

Can.Pen.Color := Grey(127); Can.MoveTo(0,4); Can.LineTo(0,Height-5);
Can.MoveTo(0,4); Can.LineTo(2,2);
Can.Pen.Width := 1; Can.MoveTo(5,7); Can.LineTo(3,9); Can.Pen.Width := 2;
Can.Pen.Color := Grey(200); Can.MoveTo(2,6); Can.LineTo(2,Height-5);//Left
Can.LineTo(4,Height-3);
Can.Pen.Color := Grey(50); Can.MoveTo(4,8); Can.LineTo(4,Height-7);

Can.Pen.Color := Grey(151); Can.MoveTo(6,Height-5); Can.LineTo(Width-5,Height-5);
Can.LineTo(Width-3,Height-7); Can.MoveTo(6,Height-5); Can.LineTo(4,Height-7);
Can.Pen.Color := Grey(224); Can.MoveTo(6,Height-3); Can.LineTo(Width-5,Height-3);//Bottom
Can.LineTo(Width-3,Height-5);
Can.Pen.Color := Grey(79); Can.MoveTo(4,Height-1); Can.LineTo(Width-3,Height-1);
Can.MoveTo(4,Height-1); Can.LineTo(2,Height-3);

Can.Pen.Color := Grey(127); Can.MoveTo(Width-5,6); Can.LineTo(Width-5,Height-5);
Can.MoveTo(Width-5,6); Can.LineTo(Width-7,4);
Can.Pen.Color := Grey(200); Can.MoveTo(Width-3,6); Can.LineTo(Width-3,Height-5);//Right
Can.Pen.Color := Grey(48); Can.MoveTo(Width-1,4); Can.LineTo(Width-1,Height-5);
Can.LineTo(Width-3,Height-3);
Can.MoveTo(Width-1,4); Can.LineTo(Width-3,2);

end;

function FourColsSame(F1,F2:TFourColours):Boolean;
begin
Result := (F1.TR=F2.TR) and (F1.TL=F2.TL) and (F1.BR=F2.BR) and (F1.BL=F2.BL);
end;

{$IFDEF BUFFERED_4_GRADIENT}
initialization
Buffer := TBitmap.Create;
finalization
Buffer.Free;
{$ENDIF}
end.
