unit lgppatch;

interface

Uses Windows, SysUtils, Classes, LGPStruct, DebugLog, FFLzs;



Const
  DEF_CREATOR: TCreator = (70,73,67,69,68,85,76,65,45,76,71,80);
  DEF_DESC: Array[0..13] of Char = ('L','G','P',' ','P','A','T','C','H',' ','F','I','L','E');

function ApplyPatch(Orig,Patch:TStream): TMemoryStream;
function PatDifferent(Orig,New:TStream): TMemoryStream;
procedure MakeLGP(Src:String;Output:String;DelFiles:Boolean);
procedure CombineArchives(Original:TRawLGPFile;Patch:TLGPFile;Destructive:Boolean);

implementation

function Min(V1,V2:Integer):Integer;
begin
If V1<V2 then Result := V1 else Result := V2;
end;
function Max(V1,V2:Integer):Integer;
begin
If V1>V2 then Result := V1 else Result := V2;
end;

function ApplyPatch(Orig,Patch:TStream): TMemoryStream;
var
Control,ControlBits,Sz: Integer;
B:                      Byte;
Buf:                    Array[1..256] of Byte;
begin
Result := TMemoryStream.Create;
Result.Size := Orig.Size;
ControlBits := 0; Sz := 0;
Patch.Position := 0; Control := $FF;
While Patch.Position < Patch.Size do begin
  If ControlBits=0 then begin
    ControlBits := 8;
    Patch.ReadBuffer(B,1);
    Control := B;
  end;

  Patch.ReadBuffer(B,1);
  If (Control and $1)=1 then begin
    Result.WriteBuffer(B,1);
    Orig.Position := Orig.Position + 1;
    Inc(Sz);
  end else begin
    Orig.ReadBuffer(Buf,B+1);
    Result.WriteBuffer(Buf,B+1);
    Inc(Sz,B+1);
  end;
  Control := Control shr 1;
  Dec(ControlBits);
end;
Result.Size := Sz;
end;

function PatDifferent(Orig,New:TStream): TMemoryStream;
var
BufO,BufN:        Array[1..256] of Byte;
ControlBits,BufLen,LastCtrl,Control,Match,I,StO,StN,Sz:   Integer;
B:                Byte;
Diff:             Boolean;
begin
Result := TMemoryStream.Create;
Result.Size := Orig.Size; Sz := 0;
StO := 0; StN := 0; ControlBits := 0;
LastCtrl := 0; Result.Position := 1;
Control := 0; Diff := False;
While (StO < Orig.Size) and (StN < New.Size) do begin
  Orig.Position := StO;
  New.Position := StN;
  BufLen := Min(Min(256,Orig.Size-StO),New.Size-StN);
  Orig.ReadBuffer(BufO,BufLen);
  New.ReadBuffer(BufN,BufLen);
  Match := 0;
  For I := 1 to BufLen do
    If BufO[I]=BufN[I] then Inc(Match) else Break;
  If Match=0 then begin
    Control := Control or (1 shl ControlBits);
    Result.WriteBuffer(BufN[1],1);
    Inc(StO); Inc(StN); Diff := True;
  end else begin
    B := Match-1;
    Result.WriteBuffer(B,1);
    Inc(StO,Match); Inc(StN,Match);
  end;
  Inc(ControlBits); Inc(Sz);
  If (ControlBits=8) then begin
    Result.Position := LastCtrl;
    B := Control;
    Result.WriteBuffer(B,1);
    Inc(Sz);
    Result.Position := Sz+1;
    LastCtrl := Sz;
    Control := 0;
    ControlBits := 0;
  end;
end;
If ControlBits<>0 then begin
    Result.Position := LastCtrl;
    B := Control;
    Result.WriteBuffer(B,1);
    Inc(Sz);
end;
Result.Size := Sz;
If Not Diff then begin
  Result.Free;
  Result := nil;
end;
end;

procedure MakeLGP(Src:String;Output:String;DelFiles:Boolean);
var
Files:      TStringList;
SR:         TSearchRec;
J,I,St:     Integer;
Data,Fil,Err:    TFileStream;
LHead:      TLGP_Header;
Tbl:        TLGP_TableEntry;
FHead:      TLGP_FileHeader;
begin
Files := TStringList.Create;

I := FindFirst(Src+'*.*',0,SR);
While I=0 do begin
  Files.Add(Src+SR.Name);
  I := FindNext(SR);
end;
FindClose(SR);

Files.Sort;

try
Err := TFileStream.Create(ExtractFilePath(ParamStr(0))+'LGP.95',fmOpenRead or fmShareDenyNone);
except
Raise Exception.Create('Couldn''t find an appropriate file validation data. Perhaps you are trying to make an LGP with an '+
            'unsupported number of files in it.');
Exit;
end;

Fil := TFileStream.Create(Output,fmCreate or fmShareExclusive);

LHead.NumFiles := Files.Count;
LHead.Creator := DEF_CREATOR;

Fil.WriteBuffer(LHead,Sizeof(LHead));

St := Sizeof(Lhead) + Sizeof(Tbl) * Files.Count + Err.Size;

For I := 0 to Files.Count-1 do begin
  ZeroMemory(@Tbl,Sizeof(Tbl));
  Tbl.Check1 := 14;
  Tbl.Check2 := 0;
  StrPCopy(Tbl.Filename,ExtractFileName(Files[I]));
  Tbl.EntryStart := St;
  Fil.WriteBuffer(Tbl,Sizeof(Tbl));
  J := FindFirst(Files[I],0,SR);
  If J<>0 then Raise Exception.Create('Couldn''t find a file that exists. AARRRRGH.');
  Inc(St, Sizeof(FHead) + SR.Size);
  FindClose(SR);
end;

Fil.CopyFrom(Err,Err.Size);
Err.Free;

For I := 0 to Files.Count-1 do begin

  ZeroMemory(@FHead,Sizeof(FHead));
  StrPCopy(FHead.Filename,ExtractFileName(Files[I]));
  Data := TFileStream.Create(Files[I],fmOpenRead or fmShareDenyNone);
  FHead.FileLen := Data.Size;
  Fil.WriteBuffer(FHead,Sizeof(FHead));
  Fil.CopyFrom(Data,Data.Size);
  Data.Free;
  If DelFiles then DeleteFile(Files[I]);
end;

Fil.WriteBuffer(DEF_DESC,Sizeof(DEF_DESC));

Fil.Free;

Files.Free;

end;

procedure CombineArchives(Original:TRawLGPFile;Patch:TLGPFile;Destructive:Boolean);
var
Tbl:    TLGP_TableEntry;
Cmp:    Boolean;
Ent:    TLGP_Entry;
J,I:    Integer;
Tmp,Nme:    String;
New,Orig,Pat:    TMemoryStream;
Raw:    TRawHeader;
begin
For I := 0 to Original.NumEntries-1 do begin
  Tbl := Original.TableEntry[I];
  Nme := Uppercase(StrPas(Tbl.Filename));
  For J := 0 to Patch.NumFiles-1 do begin
    Ent := Patch[J];
    If Uppercase(Ent.Filename)=Nme then Break;
  end;
  If Uppercase(Ent.Filename)<>Nme then Log('No matching file in patch for '+Nme) else begin
    Pat := TMemoryStream.Create;
    Pat.Size := Ent.FileLen;
    Pat.Position := 0;
    Patch.Extract(J,Pat);
    Orig := TMemoryStream.Create;
    Orig.Size := 1;
    Original.Extract(I,Orig);
    Orig.Position := 0;
    Orig.ReadBuffer(J,4);
    If J=Orig.Size-4 then begin
      New := lzs_DecompressS(Orig);
      Orig.Free;
      Orig := New;
      Cmp := True;
    end else Cmp := False;
    Pat.Position := 0;
    Pat.ReadBuffer(J,4);
    If J=Pat.Size-4 then begin
      New := lzs_DecompressS(Pat);
      Pat.Free;
      Pat := New;
      Cmp := True;
    end;
    New := ApplyPatch(Orig,Pat);
    Orig.Free;
    Pat.Free;
    If Cmp then begin
      Pat := _DodgyCompressS(New);
      New.Free;
      New := Pat;
    end;
    For J := 0 to Original.NumHeaders-1 do begin
      Raw := Original.Header[J];
      Tmp := StrPas(Raw.Head.Filename);
      Tmp := Tmp + '.patch';
      StrPCopy(Raw.Head.Filename,Tmp);
      If Raw.Offset=Tbl.EntryStart then begin
        Raw.Data := New.Memory;
        Raw.Head.FileLen := New.Size;
        If (Not Destructive) or (Not Original.ReplaceData(J,Raw)) then begin
          Original.AddHeader(Raw);
          Tbl.EntryStart := Original.Header[Original.NumHeaders-1].Offset;
          Original.TableEntry[I] := Tbl;
        end;
        Break;
      end;
    end;
    New.Free;
  end;
end;

end;

end.
