unit lgpstruct;

interface

{___$DEFINE SCANERR}

Uses Classes, Windows, SysUtils, FiceStream;

Const
  LGP_LEN_UNKNOWN = -1;
  LGP_LEN_INVALID = -2;
  LGP_UNDERRUN = $5f6a04e8;

Type

  TCreator = Array[0..11] of Byte;

  TProgressCallback = procedure(Cur,Max:Integer);
  TMessageCallback = procedure(Msg:String);

  TLGP_Header = packed record
    Creator:    TCreator;
    NumFiles:   Longint;
  end;

  TLGP_TableEntry = packed record
    Filename:   Array[0..19] of Char;
    EntryStart: Longint;
    Check1:     Byte;
    Check2:     Word;
  end;

  TLGP_FileHeader = packed record
    Filename:   Array[0..19] of Char;
    FileLen:    Longint;
  end;

  TLGP_Entry = record
    Filename:   ShortString;
    TableStart: Integer; //filestart is TableStart +24
    FileLen:    Integer;
    Error:      Boolean;
  end;

  PLGP_Entry = ^TLGP_Entry;

  TLGPFile = class
    private
      FNumFiles:  Integer;
      FList:      TList;
      SList:      TList;
      FName,FCreat,FDesc:String;
      function GetEntry(Idx:Integer): TLGP_Entry;
      function GetEntryByName(Nme:String): TLGP_Entry;
      function GetIndexByName(Nme:String): Integer;
      function GetSort(Idx:Integer): TLGP_Entry;
      procedure LoadFromStream(Src:TStream;ReadHeaders:Boolean);
    public
      property Entry[Index:Integer]: TLGP_Entry read GetEntry; default;
      property Sorted[Index:Integer]: TLGP_Entry read GetSort;
      property EntryName[Nme:String]: TLGP_Entry read GetEntryByName;
      property EntryIndex[Nme:String]: Integer read GetIndexByName;
      property NumFiles: Integer read FNumFiles;
      property Creator: String read FCreat;
      property Description: String read FDesc;
      property Filename: String read FName;
//      property Errors: Integer read FErr;
      procedure Extract(EntryNum:Integer;Dst:TStream);
      Constructor CreateFromFile(FName:String;ReadHeaders:Boolean);
      Constructor CreateFromStream(Src:TStream;ReadHeaders:Boolean);
      Destructor Destroy; Override;
  end;

  TRawHeader = packed record
    Head:     TLGP_FileHeader;
    Offset:   Integer; //pos in file
    Data:     Pointer;
    Available:Integer;
  end;

  PRawHeader = ^TRawHeader;

  PTable = ^TLGP_TableEntry;

  TRawLGPFile = class
    private
      Headers:    TList;
      Table:      TList;
      FNumHead:   Integer;
      FNumTable:  Integer;
      FCreat,FDesc:String;
      CRC:        PChar;
      Fil:        TFileStream;
      SrcFil:     String;
      function GetHeader(Index:Integer): TRawHeader;
      function GetTbl(Index:INteger): TLGP_TableEntry;
      procedure SetTbl(Index:Integer; Tbl:TLGP_TableEntry);
    public
      property Description: String read FDesc;
      property Header[Index:Integer]: TRawHeader read GetHeader;
      property TableEntry[Index:Integer]: TLGP_TableEntry read GetTbl write SetTbl;
      property NumHeaders: Integer read FNumHead;
      property NumEntries: Integer read FNumTable;
      property Creator:String read FCreat;
      property SourceFile:String read SrcFil;
      procedure AddHeader(NewHead:TRawHeader);
      function ReplaceData(HeadIndex:Integer;NewHead:TRawHeader): Boolean;
      function UpdateData(HeadIndex:Integer;Data:TMemoryStream):Boolean;
      function UpdateFile(TableIndex:Integer;Data:TMemoryStream):Boolean;
      procedure TruncateFrom(HeadNum:Integer);//zero based
      Constructor CreateFromFile(FName:String);
      Destructor Destroy; Override;
      procedure Extract(EntryNum:Integer;Dst:TStream);
  end;

var
  Progress: TProgressCallback = nil;
  Messager: TMessageCallback = nil;

Const
  DEF_CREATOR: TCreator = (0,0,83,81,85,65,82,69,83,79,70,84);
  DEF_DESC: Array[0..13] of Char = ('F','I','N','A','L',' ','F','A','N','T','A','S','Y','7');

procedure Extract(Strm:TStream;Ent:TLGP_Entry;Output:String);
procedure ExtractS(Strm:TStream;Ent:TLGP_Entry;Output:TStream);
procedure LGPRepairArchive(FName:String);
procedure LGPCreateArchive(SrcFolder,OutputFile:String;DeleteOriginal,UseErrorCheck:Boolean);
procedure LGPPackArchive(FName:String);

Const
  LGPS = 1000;

implementation

function Min(V1,V2:Integer):Integer;
begin
If V1<V2 then Result := V1 else Result := V2;
end;
function Max(V1,V2:Integer):Integer;
begin
If V1>V2 then Result := V1 else Result := V2;
end;

procedure Extract(Strm:TStream;Ent:TLGP_Entry;Output:String);
var
FHead:      TLGP_FileHeader;
Fil:        TFileStream;
begin
If Ent.FileLen=LGP_LEN_INVALID then Raise Exception.Create('LGP TOC has invalid offset!');
If Ent.FileLen=LGP_LEN_UNKNOWN then begin
  Strm.Position := Ent.TableStart;
  Strm.ReadBuffer(FHead,Sizeof(FHead));
  Ent.FileLen := FHead.FileLen;
end else Strm.Position := Ent.TableStart + Sizeof(FHead);
Fil := TFileStream.Create(Output,fmCreate or fmShareExclusive);
Fil.CopyFrom(Strm,Ent.FileLen);
Fil.Free;
end;

procedure ExtractS(Strm:TStream;Ent:TLGP_Entry;Output:TStream);
var
FHead:      TLGP_FileHeader;
begin
If Ent.FileLen=LGP_LEN_INVALID then Raise Exception.Create('LGP TOC has invalid offset!');
If Ent.FileLen=LGP_LEN_UNKNOWN then begin
  Strm.Position := Ent.TableStart;
  Strm.ReadBuffer(FHead,Sizeof(FHead));
  Ent.FileLen := FHead.FileLen;
end else Strm.Position := Ent.TableStart + Sizeof(FHead);
Output.Position := 0;
Output.CopyFrom(Strm,Ent.FileLen);
end;

procedure TLGPFile.Extract(EntryNum:Integer;Dst:TStream);
var
Fil:    TFileStream;
Head:   TLGP_FileHeader;
Ent:    TLGP_Entry;
begin
Ent := Entry[EntryNum];
Fil := TFileStream.Create(FName,fmOpenRead or fmShareDenyNone);
Fil.Position := Ent.TableStart;
Fil.ReadBuffer(Head,Sizeof(Head));
Dst.CopyFrom(Fil,Head.FileLen);
end;

Constructor TLGPFile.CreateFromStream(Src:TStream;ReadHeaders:Boolean);
begin
Inherited;
LoadFromStream(Src,ReadHeaders);
end;

Constructor TLGPFile.CreateFromFile(FName:String;ReadHeaders:Boolean);
var
Fil:    TFileStream;
begin
Inherited;
Fil := TFileStream.Create(FName,fmOpenRead or fmShareDenyNone);
LoadFromStream(Fil,ReadHeaders);
Fil.Free;
Self.FName := FName;
end;

Destructor TLGPFile.Destroy;
var
Ent:      PLGP_Entry;
I:        Integer;
begin
For I := FList.Count-1 downto 0 do begin
  Ent := FList[I];
  Dispose(Ent);
end;
FList.Free;
SList.Free;
Inherited;
end;

procedure TLGPFile.LoadFromStream(Src:TStream;ReadHeaders:Boolean);
var
Ent:        PLGP_Entry;
LHead:      TLGP_Header;
Tbl:        TLGP_TableEntry;
FHead:      TLGP_FileHeader;
Err:        TFileStream;
Mx,Lrg,I,J: Integer;
S:          String;
Ch:         Char;
Done:       Boolean;
begin
FList := TList.Create; SList := TList.Create;
Src.ReadBuffer(LHead,Sizeof(LHead));
FNumFiles := LHead.NumFiles;
FCreat := ''; FDesc := '';
For I := 0 to 11 do
  If LHead.Creator[I] <> 0 then begin
    For J := I to 11 do
      FCreat := FCreat + Char(LHead.Creator[J]);
    Break;
  end;
Lrg := 0;
Mx := FNumFiles;
If ReadHeaders then Mx := 2*Mx;
For I := 1 to FNumFiles do begin
  Src.ReadBuffer(Tbl,Sizeof(Tbl));
  New(Ent);
  Ent^.FileLen := LGP_LEN_UNKNOWN;
  Ent^.Error := False;
  Ent^.Filename := StrPas(Tbl.Filename);
  Ent^.TableStart := Tbl.EntryStart;
//  If Tbl.Check1 <> 14 then Raise Exception.Create('Checkcode failure in LGP');
//  If Tbl.Check2 <> 0 then Raise Exception.Create('Checkcode failure in LGP');
  FList.Add(Ent); Done := False;
  For J := SList.Count-1 downto 0 do
    If PLGP_Entry(SList[J]).TableStart < Ent^.TableStart then begin
      SList.Insert(J+1,Ent); Done := True;
      Break;
    end;
  If Not Done then SList.Insert(0,Ent);
  Lrg := Max(Lrg,Tbl.EntryStart);
  If @Progress<>nil then Progress(I,Mx);
end;

{$IFDEF SCANERR}
If Not FileExists(ExtractFilePath(ParamStr(0))+'LGP.'+IntToStr(FNumFiles)) then begin
  Err := TFileStream.Create(ExtractFilePath(ParamStr(0))+'LGP.'+IntToStr(FNumFiles),fmCreate or fmShareExclusive);
  Err.CopyFrom(Src,3602);
  Err.Free;
end;
{$ENDIF}

If ReadHeaders then
  For I := 0 to FList.Count-1 do begin
    Ent := FList[I];
    If Ent^.TableStart+Sizeof(FHead) > Src.Size then begin
      Ent^.FileLen := LGP_LEN_INVALID;
      Continue;
    end;
    Src.Position := Ent^.TableStart;
    Src.ReadBuffer(FHead,Sizeof(FHead));
    S := StrPas(FHead.Filename);
    If Uppercase(S) <> Uppercase(Ent^.Filename) then Ent^.Error := True;
    Ent^.FileLen := FHead.FileLen;
    If Ent^.TableStart=Lrg then begin
      Src.Position := Ent^.TableStart + Sizeof(FHead) + Ent^.FileLen;
      While (Src.Position < Src.Size) and (Length(FDesc)<255) do begin
        Src.ReadBuffer(Ch,1);
        FDesc := FDesc + Ch;
      end;
    end;
    If @Progress<>nil then Progress(FNumFiles+I,Mx);
  end;

end;

function TLGPFile.GetEntry(Idx:Integer): TLGP_Entry;
var
Ent:      PLGP_Entry;
begin
Ent := FList[Idx];
Result := Ent^;
end;

function TLGPFile.GetEntryByName(Nme:String): TLGP_Entry;
var
I:Integer;
Ent:TLGP_Entry;
begin
For I := 0 to FNumFiles-1 do begin
  Ent := GetEntry(I);
  If Uppercase(Ent.Filename)=Uppercase(Nme) then begin
    Result := Ent;
    Exit;
  end;
end;
end;

function TLGPFile.GetIndexByName(Nme:String): Integer;
var
I:Integer;
Ent:TLGP_Entry;
begin
Result := -1;
For I := 0 to FNumFiles-1 do begin
  Ent := GetEntry(I);
  If Uppercase(Ent.Filename)=Uppercase(Nme) then begin
    Result := I;
    Exit;
  end;
end;
end;

function TLGPFile.GetSort(Idx:Integer): TLGP_Entry;
var
Ent:      PLGP_Entry;
begin
Ent := SList[Idx];
Result := Ent^;
end;

Constructor TRawLGPFile.CreateFromFile(FName:String);
var
I,Sml,J:    Integer;
LHead:      TLGP_Header;
Tbl:        PTable;
Raw:        PRawHeader;
Ch:         Char;
Chk:        Array[0..1] of Integer;
begin
Inherited;
Headers := TList.Create; Table := TList.Create; CRC := nil;
FNumTable := 0; FNumHead := 0;
SrcFil := FName;
Fil := TFileStream.Create(FName,fmOpenReadWrite or fmShareDenyWrite);
Fil.ReadBuffer(LHead,Sizeof(LHead));
For I := 0 to 11 do
  If LHead.Creator[I] <> 0 then begin
    For J := I to 11 do
      FCreat := FCreat + Char(LHead.Creator[J]);
    Break;
  end;
FNumTable := LHead.NumFiles;

Sml := Fil.Size;
For I := 0 to LHead.NumFiles-1 do begin
  New(Tbl);
  Fil.ReadBuffer(Tbl^,Sizeof(TLGP_TableEntry));
  Table.Add(Tbl);
  Sml := Min(Sml,Tbl^.EntryStart);
end;

If Sml > Fil.Position then begin
  GetMem(CRC,Sml-Fil.Position);
  Fil.ReadBuffer(CRC^,Sml-Fil.Position);
end;

Raw := nil;

While (Fil.Size - Fil.Position) > Sizeof(TLGP_FileHeader) do begin
  Fil.ReadBuffer(Chk,8);
  If (Chk[0] = LGP_UNDERRUN) and (Chk[1] <> 0) then begin
    Fil.Position := Fil.Position + Chk[1] - 16;
    If Raw<>nil then Raw^.Available := Raw^.Head.FileLen+Chk[1];
  end else begin
    Fil.Seek(-8,soFromCurrent);
    If Raw<>nil then Raw^.Available := Raw^.Head.FileLen;
  end;
  New(Raw);
  Raw^.Offset := Fil.Position;
  Raw^.Data := nil;
  Fil.ReadBuffer(Raw^.Head,Sizeof(TLGP_FileHeader));
  Fil.Position := Fil.Position + Raw^.Head.FileLen;
  Headers.Add(Raw);
end;
If Raw<>nil then Raw^.Available := High(Integer);
FNumHead := Headers.Count;
FDesc := '';
While Fil.Position < Fil.Size do begin
  Fil.ReadBuffer(Ch,1);
  FDesc := FDesc + Ch;
end;

end;

Destructor TRawLGPFile.Destroy;
var
Tbl:        PTable;
Raw:        PRawHeader;
I:          Integer;
begin
For I := Headers.Count-1 downto 0 do begin
  Raw := Headers[I];
  Dispose(Raw);
end;
For I := Table.Count-1 downto 0 do begin
  Tbl := Table[I];
  Dispose(Tbl);
end;

If CRC<>nil then FreeMem(CRC);

Headers.Free; Table.Free; Fil.Free;
Inherited;
end;

function TRawLGPFile.ReplaceData(HeadIndex:Integer;NewHead:TRawHeader): Boolean;
var
Raw:      PRawHeader;
J,I:      Integer;
Ch:       Char;
begin
Result := False;
Raw := Headers[HeadIndex];
If (NewHead.Head.FileLen <> Raw^.Head.FileLen) and (NewHead.Head.FileLen > (Raw^.Available-8)) and (HeadIndex < FNumHead-1) then Exit;
Fil.Position := Raw^.Offset;
Raw^.Head.Filename := NewHead.Head.Filename;
I := Raw^.Available - NewHead.Head.FileLen;
Raw^.Head.FileLen := NewHead.Head.FileLen;
Fil.WriteBuffer(Raw^.Head,Sizeof(TLGP_FileHeader));
Fil.WriteBuffer(NewHead.Data^,Raw^.Head.FileLen);
If (I<>0) then
  If HeadIndex < FNumHead-1 then begin
    J := LGP_UNDERRUN;
    Fil.WriteBuffer(J,4);
    Fil.WriteBuffer(I,4);
  end else
    For I := 1 to Length(FDesc) do begin
      Ch := FDesc[I];
      Fil.WriteBuffer(Ch,1);
    end;

Result := True;
end;

function TRawLGPFile.UpdateFile(TableIndex:Integer;Data:TMemoryStream):Boolean;
var
I:    Integer;
Ent:  TLGP_TableEntry;
begin
Ent := TableEntry[TableIndex];
Result := False;
For I := 0 to FNumHead-1 do
  If Header[I].Offset=Ent.EntryStart then begin
    Result := UpdateData(I,Data); Exit;
  end;
end;

function TRawLGPFile.UpdateData(HeadIndex:Integer;Data:TMemoryStream):Boolean;
var
Raw:      TRawHeader;
Tbl:      TLGP_TableEntry;
Old,I:    Integer;
begin
Result := False;
If (HeadIndex<0) or (HeadIndex >= FNumHead) then Exit;
Raw := PRawHeader(Headers[HeadIndex])^;
Raw.Data := Data.Memory;
Raw.Head.FileLen := Data.Size;
If ReplaceData(HeadIndex,Raw) then begin
  Result := True; Exit;
end;
AddHeader(Raw);
Old := Raw.Offset;
Raw := PRawHeader(Headers[FNumHead-1])^;
For I := 0 to FNumTable-1 do begin
  Tbl := TableEntry[I];
  If Tbl.EntryStart=Old then begin
    Tbl.EntryStart := Raw.Offset;
    TableEntry[I] := Tbl;
  end;
end;
Result := True;
end;

procedure TRawLGPFile.AddHeader(NewHead:TRawHeader);
var
Raw:      PRawHeader;
Ch:       Char;
I:        Integer;
begin
Raw := Headers[Headers.Count-1];
Raw^.Available := Raw^.Head.FileLen;
Fil.Position := Raw^.Offset + Raw^.Head.FileLen + Sizeof(TLGP_FileHeader);
NewHead.Offset := Fil.Position;
Fil.WriteBuffer(NewHead.Head, Sizeof(TLGP_FileHeader));
Fil.WriteBuffer(NewHead.Data^, NewHead.Head.FileLen);
For I := 1 to Length(FDesc) do begin
  Ch := FDesc[I];
  Fil.WriteBuffer(Ch,1);
end;
New(Raw);
Raw^ := NewHead;
Raw^.Available := High(Integer);
Headers.Add(Raw);
Inc(FNumHead);
end;

function TRawLGPFile.GetHeader(Index:Integer): TRawHeader;
var
Raw:    PRawHeader;
begin
Raw := Headers[Index];
Result := Raw^;
end;

function TRawLGPFile.GetTbl(Index:Integer): TLGP_TableEntry;
var
Tbl:        PTable;
begin
Tbl := Table[Index];
Result := Tbl^;
end;

procedure TRawLGPFile.Extract(EntryNum:Integer;Dst:TStream);
var
Head:   TLGP_FileHeader;
Ent:    TLGP_TableEntry;
begin
Ent := TableEntry[EntryNum];
Fil.Position := Ent.EntryStart;
Fil.ReadBuffer(Head,Sizeof(Head));
Dst.CopyFrom(Fil,Head.FileLen);
end;

procedure TRawLGPFile.SetTbl(Index:Integer; Tbl:TLGP_TableEntry);
var
ETbl:        PTable;
begin
ETbl := Table[Index];
ETbl^ := Tbl;
Fil.Position := Sizeof(TLGP_Header) + Sizeof(TLGP_TableEntry) * Index;
Fil.WriteBuffer(Tbl,Sizeof(Tbl));
end;

procedure TRawLGPFile.TruncateFrom(HeadNum:Integer);//zero based
var
I:  Integer;
Ch: Char;
Head:TRawHeader;
begin
Head := Header[HeadNum];
Fil.Size := Head.Offset + Length(FDesc);
Fil.Position := Head.Offset;
For I := 1 to Length(FDesc) do begin
  Ch := FDesc[I];
  Fil.WriteBuffer(Ch,1);
end;
For I := FNumHead-1 downto HeadNum do
  Headers.Delete(I);
FNumHead := HeadNum;
end;

procedure LGPRepairArchive(FName:String);
var
Raw:      TRawLGPFile;
I,J:      Integer;
NewHead:  TRawHeader;
Ent:      TLGP_TableEntry;
begin
Raw := TRawLGPFile.CreateFromFile(FName);
If @Progress<>nil then Progress(0,Raw.NumEntries);
For I := 0 to Raw.NumEntries-1 do begin
  Ent := Raw.TableEntry[I];
  For J := 0 to Raw.NumHeaders-1 do begin
    NewHead := Raw.Header[J];
    If Uppercase(Ent.Filename) = Uppercase(StrPas(NewHead.Head.Filename)) then begin
      Ent.EntryStart := NewHead.Offset;
      Raw.TableEntry[I] := Ent;
      Break;
    end;
  end;
  If @Progress<>nil then Progress(I+1,Raw.NumEntries);
end;
Raw.Free;

end;

procedure LGPCreateArchive(SrcFolder,OutputFile:String;DeleteOriginal,UseErrorCheck:Boolean);
var
Files:        TStringList;
SR:           TSearchRec;
J,I,St:       Integer;
Data,Fil:     TFileStream;
Err:          TStream;
LHead:        TLGP_Header;
Tbl:          TLGP_TableEntry;
FHead:        TLGP_FileHeader;
begin
Files := TStringList.Create;
If SrcFolder[Length(SrcFolder)]<>'\' then SrcFolder := SrcFolder + '\';

I := FindFirst(srcFolder+'*.*',0,SR);
While I=0 do begin
  Files.Add(srcFolder+SR.Name);
  I := FindNext(SR);
end;
FindClose(SR);

Files.Sort;

If UseErrorCheck then
try
//Err := TFileStream.Create(ExtractFilePath(ParamStr(0))+'LGP.'+IntToStr(Files.Count),fmOpenRead or fmShareDenyNone);
Err := FiceInputStream('LGP_'+IntToStr(Files.Count));
If Err=nil then Raise EFiceException.Create('No validation data!');
except
If @Messager<>nil then
  Messager('Couldn''t find an appropriate file validation data. Perhaps you are trying to make an LGP with an '+
            'unsupported number of files in it.');
Exit;
end;

Fil := TFileStream.Create(OutputFile,fmCreate or fmShareExclusive);

LHead.NumFiles := Files.Count;
LHead.Creator := DEF_CREATOR;

Fil.WriteBuffer(LHead,Sizeof(LHead));

St := Sizeof(Lhead) + Sizeof(Tbl) * Files.Count;
If UseErrorCheck then Inc(St,Err.Size);

For I := 0 to Files.Count-1 do begin
  ZeroMemory(@Tbl,Sizeof(Tbl));
  Tbl.Check1 := 14;
  Tbl.Check2 := 0;
  StrPCopy(Tbl.Filename,ExtractFileName(Files[I]));
  Tbl.EntryStart := St;
  Fil.WriteBuffer(Tbl,Sizeof(Tbl));
  J := FindFirst(Files[I],0,SR);
  If J<>0 then Raise Exception.Create('Couldn''t find a file that exists. AARRRRGH.');
  Inc(St, Sizeof(FHead) + SR.Size);
  FindClose(SR);
end;

If UseErrorCheck then begin
  Fil.CopyFrom(Err,Err.Size);
  Err.Free;
end;

If @Progress<>nil then Progress(0,Files.Count);

For I := 0 to Files.Count-1 do begin

  ZeroMemory(@FHead,Sizeof(FHead));
  StrPCopy(FHead.Filename,ExtractFileName(Files[I]));
  Data := TFileStream.Create(Files[I],fmOpenRead or fmShareDenyNone);
  FHead.FileLen := Data.Size;
  Fil.WriteBuffer(FHead,Sizeof(FHead));
  Fil.CopyFrom(Data,Data.Size);
  Data.Free;
  If DeleteOriginal then DeleteFile(Files[I]);
  If @Progress<>nil then Progress(0,I+1);
end;

Fil.WriteBuffer(DEF_DESC,Sizeof(DEF_DESC));

Fil.Free;

Files.Free;

end;

procedure LGPPackArchive(FName:String);
var
CurOS:    Integer;
Raw:      TRawLGPFile;
I,J,K:    Integer;
Used:     Boolean;
Head:     TRawHeader;
Tmp:      TMemoryStream;
Tbl:      TLGP_TableEntry;
Ch:       Char;
begin
Raw := TRawLGPFile.CreateFromFile(FName);
CurOS := High(Integer);
For I := 0 to Raw.FNumHead-1 do
  CurOS := Min(Raw.Header[I].Offset,CurOS);

Tmp := TMemoryStream.Create;

For J := 0 to Raw.FNumHead-1 do begin
  If @Progress<>nil then Progress(J,Raw.FNumHead);
  Used := False;
  Head := Raw.Header[J];
  For K := 0 to Raw.FNumTable-1 do
    If Raw.TableEntry[K].EntryStart = Head.Offset then begin
      Used := True; Break;
    end;
  If Not Used then Continue;
  Tmp.Clear;
  Raw.Fil.Position := Head.Offset;
  Tmp.CopyFrom(Raw.Fil,Head.Head.FileLen + Sizeof(Head.Head));
  Raw.Fil.Position := CurOS;
  Raw.Fil.CopyFrom(Tmp,0);
  For K := 0 to Raw.FNumTable-1 do begin
    Tbl := Raw.TableEntry[K];
    If Tbl.EntryStart = Head.Offset then begin
      Tbl.EntryStart := CurOS;
      Raw.TableEntry[K] := Tbl;
    end;
  end;
  Inc(CurOS,Head.Head.FileLen + Sizeof(Head.Head));
end;

Raw.Fil.Position := CurOS;
For I := 1 to Length(Raw.FDesc) do begin
  Ch := Raw.FDesc[I];
  Raw.Fil.WriteBuffer(Ch,1);
end;
Raw.Fil.Size := CurOS + Length(Raw.FDesc);

If @Progress<>nil then Progress(Raw.FNumHead,Raw.FNumHead);
Raw.Free;
Tmp.Free;
end;

end.
