unit ff73d;

interface

Uses Classes, Windows, Cappo, SysUtils, Dialogs, GL;

Type

  TPoint3d = packed record
    Case Boolean of
    True: (
      Data: array[0..2] of Single;
    );
    False: (
      X,Y,Z: Single;
    );
    end;

  PPoint3d = ^TPoint3d;

  TFF7PolyLine = packed record
    vertSrc,vertDest:     Word;
  end;
  TFF7Polygon = packed record
    Tag:                  Word;
    Verts,CVerts,Lines:   Array[0..2] of Word;
    Unknown:              Longint;
  end;

  PFF7Polygon = ^TFF7Polygon;

  TFF7PPart = class
    private
      Verts, Colours:     PPoint3d;
      NumVert,NumColour:  Integer;
      Polys:              PFF7Polygon;
      NumPoly:            Integer;
      procedure Clearup;
    public
      Parent:     Integer;
      Offset:     Single;
      Low,High:   Single;
      Rot:        TPoint3d;
      Name:       String;
      Constructor Create;
      Constructor CreateFromFile(PFile:String);
      Constructor CreateFromStream(PStream:TStream);
      procedure LoadFromFile(PFile:String);
      procedure LoadFromStream(PStream:TStream);
      Destructor Destroy; Override;
      procedure RenderGL;
  end;

function Point3d(X,Y,Z:Single):TPoint3d;
function PointsEqual(P1,P2:TPoint3d):Boolean;
function AddPoints(P1,P2:TPoint3d):TPoint3d;
procedure ParseGeometry(PStream:TMemoryStream);


var
Vertices:     Array[0..999] of TPoint3d;
NumVerts:     Integer;
Colours:      Array[0..999] of TPoint3d;
NumColours:   Integer;
Polygons:     Array[0..999] of TFF7Polygon;
NumPolys:     Integer;
ColUsed:      Array[0..999] of Boolean;
Used:         Integer;
DefColours:   Array[0..9] of TPoint3d;
ZREV:         Boolean=True;
implementation

function Point3d(X,Y,Z:Single):TPoint3d;
begin
Result.X := X;
Result.Y := Y;
Result.Z := Z;
end;

function AddPoints(P1,P2:TPoint3d):TPoint3d;
begin
Result.X := P1.X+P2.X;
Result.Y := P1.Y+P2.Y;
Result.Z := P1.Z+P2.Z;
end;

function PointsEqual(P1,P2:TPoint3d):Boolean;
begin
Result := (P1.X=P2.X) and (P1.Y=P2.Y) and (P1.Z=P2.Z);
end;

procedure ParseGeometry(PStream:TMemoryStream);
var
ITmp,J,I,Some1,Some2,Some3,nVerts,nLines,Maxm:       Integer;
Tmp:                                              Array[1..4] of Byte;
Quad: TRGBQuad;
begin
Maxm := 0; UseD:= 0;

PStream.Position := $10; PStream.ReadBuffer(Some1,4);
PStream.Position := $0C; PStream.ReadBuffer(Some2,4);
PStream.Position := $18; PStream.ReadBuffer(Some3,4);
PStream.Position := $20; PStream.ReadBuffer(nLines,4);
PStream.Position := $24; PStream.ReadBuffer(nVerts,4);

For I := 0 to Some1+Some2-1 do begin
  PStream.Position := $80 + 12*I;
  PStream.ReadBuffer(Vertices[I],Sizeof(TPoint3d));
  Vertices[I].Y := -Vertices[I].Y;
  Vertices[I].Z := -Vertices[I].Z;
end;
NumVerts := Some1+Some2;

NumColours := nVerts+Some2;
  PStream.Position := $80 + (Some1+Some2)*12 + Some3*8;
  PStream.ReadBuffer(Quad,4);

For I := 0 to NumColours-1 do begin
  PStream.Position :=  $80 + (Some1+Some2)*12 + Some3*8 + I*Sizeof(Tmp);
//  PStream.Position := 3428-4+I*Sizeof(Tmp);
  PStream.ReadBuffer(Quad,Sizeof(Quad));
  ColUsed[I] := False;
  Colours[I].X := (Quad.rgbRed)/ 205;
  Colours[I].Y := (Quad.rgbGreen)/ 205;
  Colours[I].Z := (Quad.rgbBlue)/ 205;
end;

For I := 0 to nVerts-1 do begin
  PStream.Position := $80 + (Some1+Some2)*12 + Some3*8 + (Some2+nVerts)*4 + nLines*4 + I*Sizeof(TFF7Polygon);
  PStream.ReadBuffer(Polygons[I],Sizeof(TFF7Polygon));
  For J := 0 to 2 do
    If Not ColUsed[Polygons[I].CVerts[J]] then begin
      ColUsed[Polygons[I].CVerts[J]]:=True; Inc(Used);
    end;
  Maxm := Max(Maxm,Max(Max(Polygons[I].CVerts[0],Polygons[I].CVerts[1]),Polygons[I].CVerts[2]));
end;
ShowMessage(IntToStr(Maxm)+'x'+IntToStr(Used)+'x'+IntToStr(NumColours));
NumPolys := nVerts;
For I := 0 to Maxm do ColUsed[I] := False;
ITmp := 0;
For I := 0 to Maxm do begin
  If ColUsed[I] then Continue;
  For J := Maxm downto I do
    If PointsEqual(Colours[I],Colours[J]) then begin
//      Colours[J] := DefColours[ITmp];
      ColUsed[J] := True;
    end;
    ITmp := Min(ITmp+1,9);
end;


end;

Constructor TFF7PPart.Create;
begin
Inherited Create;
Verts := nil; Colours := nil; Polys := nil;
NumVert := 0; NumColour := 0; NumPoly := 0;
Low := 0; High := 0;
end;


Constructor TFF7PPart.CreateFromFile(PFile:String);
begin
Inherited Create;
Verts := nil; Colours := nil; Polys := nil;
NumVert := 0; NumColour := 0; NumPoly := 0;
Low := 0; High := 0;
LoadFromFile(PFile);
end;

Constructor TFF7PPart.CreateFromStream(PStream:TStream);
begin
Inherited Create;
Verts := nil; Colours := nil; Polys := nil;
NumVert := 0; NumColour := 0; NumPoly := 0;
Low := 0; High := 0;
LoadFromStream(PStream);
end;

procedure TFF7PPart.LoadFromFile(PFile:String);
var
Fil:    TFileStream;
begin
Fil := TFileStream.Create(PFile,fmOpenRead or fmShareDenyNone);
LoadFromStream(Fil);
Fil.Free;
end;

procedure TFF7PPart.LoadFromStream(PStream:TStream);
var
ITmp,J,I,Some1,Some2,Some3,nVerts,nLines,Maxm:       Integer;
TmpPt:    PPoint3d;
TmpPoly:  PFF7Polygon;
Quad: TRGBQuad;
begin
Clearup;

PStream.Position := $10; PStream.ReadBuffer(Some1,4);
PStream.Position := $0C; PStream.ReadBuffer(Some2,4);
PStream.Position := $18; PStream.ReadBuffer(Some3,4);
PStream.Position := $20; PStream.ReadBuffer(nLines,4);
PStream.Position := $24; PStream.ReadBuffer(nVerts,4);

NumVert := Some1+Some2;
GetMem(Verts,Sizeof(TPoint3d)*NumVert);
TmpPt := Verts;
//If Not ZREV then ShowMessage('Not reversing!');
For I := 0 to NumVert-1 do begin
  PStream.Position := $80 + 12*I;
  PStream.ReadBuffer(TmpPt^,Sizeof(TPoint3d));
//  If ZREV then TmpPt^.X := -TmpPt^.X;
  If ZREV then TmpPt^.Y := -TmpPt^.Y;
  If ZREV then TmpPt^.Z := -TmpPt^.Z;
  If I=0 then begin
    Low := TmpPt^.Z; High := TmpPt^.Z;
  end else begin
    Low := MinF(TmpPt^.Z,Low); High := MaxF(TmpPt^.Z,High);
  end;
  Inc(TmpPt);
end;

NumColour := nVerts+Some2;
GetMem(Colours,Sizeof(TPoint3d)*NumColour);
TmpPt := Colours;
For I := 0 to NumColour-1 do begin
  PStream.Position := $80 + (Some1+Some2)*12 + Some3*8 + I*Sizeof(Quad);
  PStream.ReadBuffer(Quad,Sizeof(Quad));
  TmpPt^.X := (Quad.rgbRed)/ 205;
  TmpPt^.Y := (Quad.rgbGreen)/ 205;
  TmpPt^.Z := (Quad.rgbBlue)/ 205;
  Inc(TmpPt);
end;

NumPoly := nVerts;
GetMem(Polys,Sizeof(TFF7Polygon)*NumPoly);
TmpPoly := Polys;
For I := 0 to NumPoly-1 do begin
  PStream.Position := $80 + (Some1+Some2)*12 + Some3*8 + (Some2+nVerts)*4 + nLines*4 + I*Sizeof(TFF7Polygon);
  PStream.ReadBuffer(TmpPoly^,Sizeof(TFF7Polygon));
  Inc(TmpPoly);
end;


end;

procedure TFF7PPart.Clearup;
begin
If Verts<>nil then FreeMem(Verts);
If Colours<>nil then FreeMem(Colours);
If Polys<>nil then FreeMem(Polys);
Verts := nil; Colours := nil; Polys := nil;
end;

procedure TFF7PPart.RenderGL;
var
I,J,N:    Integer;
TmpPt:    PPoint3d;
TmpPoly:  PFF7Polygon;
begin
glBegin(GL_LINES);
glVertex3f(-25,0,0);
glVertex3f(25,0,0);
glVertex3f(0,-25,0);
glVertex3f(0,25,0);
glVertex3f(0,0,-25);
glVertex3f(0,0,25);
glEnd;

glBegin(GL_TRIANGLES);
  TmpPoly := Polys;
  For I := 1 to NumPoly do begin
    For J := 0 to 2 do begin
      N := TmpPoly^.Verts[J];
      TmpPt := Colours;
      Inc(TmpPt,N);
//      If N<NumColour then
      glColor3fv(Pointer(TmpPt));
      TmpPt := Verts;
      Inc(TmpPt,N);
//      If N<NumVert then
      glVertex3fv(Pointer(TmpPt));
    end;
    Inc(TmpPoly);
  end;
glEnd;

end;

Destructor TFF7PPart.Destroy;
begin
Clearup;
Inherited Destroy;
end;

initialization
DefColours[0] := Point3d(1,0,0);
DefColours[1] := Point3d(0,1,0);
DefColours[2] := Point3d(0,0,1);
DefColours[3] := Point3d(1,1,0);
DefColours[4] := Point3d(1,0,1);
DefColours[5] := Point3d(0,1,1);
DefColours[6] := Point3d(1,1,1);
DefColours[7] := Point3d(0,0,0);
DefColours[8] := Point3d(0.5,0.5,0.5);
DefColours[9] := Point3d(0.5,0.5,0);

end.
