unit ffctrl;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  KapGrad, Cappo, MPlayer;

type

  TFFHand = class(TGraphicControl)
  protected
           procedure Paint; Override;
           procedure SetHand(Val:TBitmap);
           procedure SetIndex(Val:Integer);
           procedure SetControl(Val:TControl);
           procedure Blip;
           procedure MakeProperList(Lst:TList);
  private
         FHand:      TBitmap;
         FPattern:   String;
         FIndex:     Integer;
         FMask:      TColor;
         FCurrent:   TControl;
         FMedia:     TMediaPlayer;
         FTagSort:   Boolean;
  public
        Constructor Create(AOwner:TComponent); Override;
        Destructor Destroy; Override;
        procedure Forwards; procedure Backwards;
        property CurControl: TControl read FCurrent write SetControl;
  published
           property Hand: TBitmap read FHand write SetHand;
           property ItemIndex: Integer read FIndex write SetIndex;
           property Pattern: String read FPattern write FPattern;
           property Transparent: TColor read FMask write FMask;
           property Blipper: TMediaPlayer read FMedia write FMedia;
           property TagSort: Boolean read FTagSort write FTagSort;
  end;

  TFFMulti = class(TGraphicControl)
  private
    { Private declarations }
  protected
    FOpts:                 TStrings;
    FSelColor,FNonColor,FShColor:   TColor;
    FIndex:                Integer;
    FSize:                 Boolean;
    { Protected declarations }
    procedure Paint; Override;
    procedure Loaded; Override;
    procedure SetFOpts(Val:TStrings);
    procedure SetIndex(Val:Integer);
    procedure SetAutoSize(Val:Boolean);
  public
    { Public declarations }
    Constructor Create(AOwner:TComponent); Override;
    Destructor Destroy; Override;
  published
    { Published declarations }
    property Font;
    property ParentFont;
    property Options: TStrings read FOpts write SetFOpts;
    property SelColor: TColor read FSelColor write FSelColor;
    property NonColor: TColor read FNonColor write FNonColor;
    property ShadowColor: TColor read FShColor write FShColor;
    property ItemIndex: Integer read FIndex write FIndex;
    property AutoHeight: Boolean read FSize write SetAutoSize;
  end;

  TFFPanel = class(TGraphicControl)
  protected
           FBorder:        Boolean;
           BMP:            TBitmap;
           procedure Paint; Override;
           procedure Loaded; Override;
  public
        Colours:          TFourColours;
        Constructor Create(AOwner:TComponent); Override;
        Destructor Destroy; Override;
        procedure Calculate;
  published
           property Border:    Boolean read FBorder write FBorder;
           property Align;
  end;

procedure Register;
function SortByTag(Item1, Item2: Pointer): Integer;

implementation

procedure Register;
begin
  RegisterComponents('Cappo', [TFFMulti,TFFPanel,TFFHand]);
end;

procedure TffMulti.Paint;
var
I,Sz:     Integer;
begin
If FOpts.Count=0 then Exit;
SetAutoSize(FSize);
Canvas.Font := Font;
Canvas.Brush.Style := bsClear;
Sz := Width div FOpts.Count;
For I := 0 to FOpts.Count-1 do begin
    Canvas.Font.Color := FShColor;
    Canvas.TextOut( (Sz*I)+2,2,FOpts[I]);
    If I=FIndex then Canvas.Font.Color := FSelColor
                else Canvas.Font.Color := FNonColor;
    Canvas.TextOut(Sz*I,0,FOpts[I]);
end;
end;

Constructor TffMulti.Create(AOwner:TComponent);
begin
Inherited Create(AOwner);
FOpts := TStringList.Create;
end;

Destructor TffMulti.Destroy;
begin
FOpts.Free;
Inherited Destroy;
end;

Procedure TffMulti.SetFOpts(Val:TStrings);
begin
FOpts.Assign(Val);
SetAutoSize(FSize);
end;

procedure TffMulti.SetIndex(Val:Integer);
begin
If (Val<-1) or (Val >= FOpts.Count) then Raise EListError.Create('Itemindex Out of Bounds!')
   else begin
   FIndex := Val;
   Repaint;
   end;
end;

procedure TffMulti.Loaded;
begin
SetAutoSize(FSize);
end;

procedure TffMulti.SetAutoSize(Val:Boolean);
var
Hgt:      Integer;
begin
FSize := Val;
If FSize then begin
   Hgt := Canvas.TextHeight('KgYy');
   If Height <> Hgt then SetBounds(Left,Top,Width,Hgt);
end;
end;

procedure TFFPanel.Paint;
begin
Canvas.StretchDraw(Rect(0,0,Width,Height),BMP);
end;

Constructor TFFPanel.Create(AOwner:TComponent);
begin
Inherited;
BMP := TBitmap.Create;
BMP.PixelFormat := pf32Bit;
Colours := ffColours;
end;

Destructor TFFPanel.Destroy;
begin
BMP.Free;
Inherited;
end;

procedure TFFPanel.Calculate;
begin
BMP.Width := Width; BMP.Height := Height;
FilOpt(BMP,Width,Height,Colours);
If FBorder then ffBorder(BMP.Canvas,Width+1,Height+1,0);
end;

procedure TFFPanel.Loaded;
begin
Calculate;
end;

Constructor TFFHand.Create(AOwner:TComponent);
begin
Inherited;
FHand := TBitmap.Create;
end;

Destructor TFFHand.Destroy;
begin
FHand.Free;
Inherited;
end;

procedure TFFHand.SetHand(Val:TBitmap);
begin
FHand.Assign(Val);
SetBounds(Left,Top,FHand.Width,FHand.Height);
end;

procedure TFFHand.Paint;
var
IList:    TImageList;
Tmp:      TBitmap;
begin
If (FHand.Width=0) or (FHand.Height=0) then Exit;
IList := TImageList.Create(nil); IList.Height := FHand.Height;
IList.Width := FHand.Width;
IList.DrawingStyle := dsNormal; IList.BkColor := clNone;
Tmp := TBitmap.Create; Tmp.Assign(FHand);
IList.AddMasked(Tmp,FMask);
IList.Overlay(0,1);
IList.DrawOverlay(Canvas,0,0,0,1);
IList.Free;
Tmp.Free;
end;

procedure TFFHand.Blip;
begin
If FMedia=nil then Exit;
If (FMedia.Mode=mpStopped) or (FMedia.Mode=mpOpen) then FMedia.Play;
end;

procedure TFFHand.SetIndex(Val:Integer);
var
Objs:     TList;
begin
If Val<1 then begin Hide; FIndex := 0; Exit; end;
Objs := TList.Create;
MakeProperList(Objs);
If Objs.Count=0 then begin FIndex := 0; Exit; Hide; end;
Show;
If Val>Objs.Count then Val := Objs.Count;
FCurrent := Objs[Val-1]; FIndex := Val;
SetBounds(FCurrent.Left - FHand.Width,FCurrent.Top + (FCurrent.Height - FHand.Height) div 2,
                    FHand.Width,FHand.Height);
Blip;
Objs.Free;
end;

procedure TFFHand.SetControl(Val:TControl);
var
I:        Integer;
Ctrl:     TControl;
Objs:     TList;
begin
If Val=nil then begin Hide; FCurrent := nil; Exit; end;
Objs := TList.Create;
MakeProperList(Objs);
For I := 0 to Objs.Count-1 do begin
    Ctrl := TControl(Objs[I]);
    If Val=Ctrl then begin
       SetBounds(Ctrl.Left - FHand.Width,Ctrl.Top + (Ctrl.Height - FHand.Height) div 2,
                 FHand.Width,FHand.Height);
       FCurrent := Ctrl;
       FIndex := I+1;
       Blip;
       Break;
    end;
end;
Objs.Free;
end;

procedure TFFHand.Forwards;
begin
SetIndex(FIndex+1);
end;

procedure TFFHand.Backwards;
begin
SetIndex(Max(FIndex-1,1));
end;

procedure TFFHand.MakeProperList(Lst:TList);
var
I:        Integer;
begin
For I := 0 to Parent.ControlCount-1 do
    If MatchWildcard(FPattern,Parent.Controls[I].Name) and
       Parent.Controls[I].Enabled then Lst.Add(Parent.Controls[I]);
If FTagSort then Lst.Sort(SortByTag);
end;

function SortByTag(Item1, Item2: Pointer): Integer;
begin
Result := TControl(Item1).Tag - TControl(Item2).Tag;
end;

end.
